"""
Functions that actually call emg3d within the CLI interface.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import os
import sys
import json
import time
import logging

import numpy as np

from emg3d import io, utils, simulations
from emg3d.cli import parser


def simulation(args_dict):
    """Run `emg3d` invoked by CLI.

    Run and log `emg3d` given the settings stored in the config file, overruled
    by settings passed in `args_dict` (which correspond to command-line
    arguments).

    Results are saved to files according to provided settings.


    Parameters
    ----------
    args_dict : dict
        Arguments from terminal, see :func:`emg3d.cli.main`. Parameters passed
        in `args_dict` overrule parameters in the `config`.

    """

    # Start timer.
    runtime = utils.Time()

    # Parse configuration file.
    cfg, term = parser.parse_config_file(args_dict)
    check_files(cfg, term)  # Check all files and directories exist.
    function, verb = term['function'], term['verbosity']
    dry_run = term.get('dry_run', False)

    # Start this task: start timing.
    logger = initiate_logger(cfg, runtime, verb)

    # Log start info, python and emg3d version, and python path.
    logger.info(f":: emg3d CLI {function} START :: {time.asctime()} :: "
                f"v{utils.__version__}")
    logger.debug(f"{utils.Report()}")

    # Dump the configuration.
    paramdump = json.dumps(cfg, sort_keys=True, indent=4)
    logger.debug("\n    :: CONFIGURATION ::\n")
    logger.debug(f"{term['config_file']}\n{paramdump}")

    # Load input.
    logger.info("\n    :: LOAD SURVEY AND MODEL ::\n")
    sdata, sinfo = io.load(cfg['files']['survey'], verb=-1)
    survey = sdata['survey']
    logger.info(sinfo.split('\n')[0])
    logger.debug(sinfo.split('\n')[1])
    model, minfo = io.load(cfg['files']['model'], verb=-1)
    logger.info(minfo.split('\n')[0])
    logger.debug(minfo.split('\n')[1])
    min_offset = cfg['simulation_options'].pop('min_offset', 0.0)

    # Select data.
    data = cfg['data']
    if data:
        survey = survey.select(sources=data.get('sources', None),
                               receivers=data.get('receivers', None),
                               frequencies=data.get('frequencies', None))

    # Create simulation.
    sim = simulations.Simulation(
            survey=survey,
            grid=model['mesh'],
            model=model['model'],
            verb=-1,  # Only errors.
            **cfg['simulation_options']
            )

    # Switch-off tqdm if verbosity is zero.
    if verb < 1:
        sim._tqdm_opts['disable'] = True

    # Print simulation info.
    logger.info("\n    :: SIMULATION ::")
    logger.info(f"\n{sim}\n")

    # Print meshes.
    logger.debug("    :: MESHES ::\n")
    logger.debug(sim.print_grid_info(return_info=True))

    # Initiate output dict, add configuration.
    # Ideally, we would add the entire configuration; however, this causes
    # currently problems for saving h5. We therefore save only the data
    # selection info, such that we know what data was selected.
    # output = {'configuration': cfg}
    output = {'configuration': {'data': cfg.get('data', {})}}

    # Compute forward model (all calls).
    logger.info("    :: FORWARD COMPUTATION ::\n")
    if dry_run:
        output['data'] = np.zeros(sim.survey.shape, dtype=complex)
    else:

        if function == 'forward':
            sim.compute(observed=True, min_offset=min_offset)
            output['data'] = sim.data.observed
        else:
            sim.compute()
            output['data'] = sim.data.synthetic

        # Print Solver Logs.
        if verb in [0, 1]:
            sim.print_solver_info('efield', 0)
        logger.debug(sim.print_solver_info('efield', 1, True))

    # Compute the misfit.
    if function in ['misfit', 'gradient']:
        if dry_run:
            output['misfit'] = 0.0
        else:
            output['misfit'] = sim.misfit
        output['n_observations'] = sim.survey.size

    # Compute the gradient.
    if function == 'gradient':
        logger.info("\n    :: BACKWARD COMPUTATION ::\n")

        if dry_run:
            output['gradient'] = np.zeros(model['mesh'].vnC)
        else:
            output['gradient'] = sim.gradient

            # Print Solver Logs.
            if verb in [0, 1]:
                sim.print_solver_info('bfield', 0)
            logger.debug(sim.print_solver_info('bfield', 1, True))

    # Store output to disk.
    logger.info("    :: SAVE RESULTS ::\n")
    if cfg['files']['store_simulation']:
        output['simulation'] = sim
    oinfo = io.save(cfg['files']['output'], **output, verb=-1)
    logger.info(oinfo.split('\n')[0])
    logger.debug(oinfo.split('\n')[1])

    # Goodbye
    logger.info(f"\n:: emg3d CLI {function} END   :: {time.asctime()} :: "
                f"runtime = {runtime.runtime}")


def check_files(cfg, term):
    """Ensure all paths and files exist."""
    error = ""

    # First check if config file exists.
    fname = term['config_file']
    if not os.path.isfile(fname) and fname != '.':  # '.' => no config file.
        error += f"* ERROR   :: Config file not found: {fname}\n"

    # Check Survey and Model.
    for name in ['Survey', 'Model']:
        fname = cfg['files'][name.lower()]
        if not os.path.isfile(fname):
            error += f"* ERROR   :: {name} file not found: {fname}\n"

    # Finally check output directory.
    dname = os.path.split(cfg['files']['log'])[0]
    if not os.path.isdir(dname):
        error += f"* ERROR   :: Output directory does not exist: {dname}\n"

    # If any was not found, exit with error.
    if len(error) > 10:
        sys.exit(error[:-1])


def initiate_logger(cfg, runtime, verb):
    """Initiate logger for CLI of emg3d."""

    # Get logger of emg3d.cli.run and add handles.
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    # Remove the corresponding handlers if they exist already
    # (e.g., consecutive runs in IPython).
    for h in logger.handlers[:]:
        if h.name in ['emg3d_fh', 'emg3d_ch']:
            logger.removeHandler(h)
        h.close()

    # Create file handler; logs everything.
    fh = logging.FileHandler(f"{cfg['files']['log']}", mode='w')
    fh.setLevel(logging.DEBUG)
    fh_format = logging.Formatter('{message}', style='{')
    fh.setFormatter(fh_format)
    fh.set_name('emg3d_fh')  # Add name to easy remove them.
    logger.addHandler(fh)

    # Create console handler.
    ch = logging.StreamHandler()
    ch.setLevel([40, 30, 20, 10][verb+1])
    ch_format = logging.Formatter('{message}', style='{')
    ch.setFormatter(ch_format)
    ch.set_name('emg3d_ch')  # Add name to easy remove them.
    logger.addHandler(ch)

    # Add handlers to Python Warnings.
    logging.captureWarnings(True)
    logger_warnings = logging.getLogger("py.warnings")
    logger_warnings.setLevel(logging.DEBUG)
    logger_warnings.addHandler(ch)
    logger_warnings.addHandler(fh)

    return logger



"""
A simulation is the computation (modelling) of electromagnetic responses of a
resistivity (conductivity) model for a given survey.

In its heart, `emg3d` is a multigrid solver for 3D electromagnetic diffusion
with tri-axial electrical anisotropy. However, it contains most functionalities
to also act as a modeller. The simulation module combines all these things
by combining surveys with computational meshes and fields and providing
high-level, specialised modelling routines.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import warnings
import itertools
from copy import deepcopy

import numpy as np
import scipy.linalg as sl

try:
    from tqdm.contrib.concurrent import process_map
except ImportError:
    # If you have tqdm installed, but don't want to use it, simply do
    # `emg3d.simulation.process_map = emg3d.utils._process_map`.
    from emg3d.utils import _process_map as process_map

from emg3d import fields, solver, surveys, maps, models, meshes, optimize

__all__ = ['Simulation', 'expand_grid_model', 'estimate_gridding_opts']


class Simulation:
    """Create a simulation for a given survey on a given model.

    A simulation can be used to compute responses for an entire survey, hence
    for an arbitrary amount of sources, receivers, and frequencies. The
    responses can be computed in parallel over sources and frequencies. It can
    also be used to compute the misfit with the data and to compute the
    gradient of the misfit function.

    The computational grid(s) can either be provided, or automatic gridding can
    be used; see the description of the parameters ``gridding`` and
    ``gridding_opts`` for more details.

    .. warning::

        The automatic gridding does its best to generate meshes that are
        suitable for the provided model and survey. However, CSEM spans a wide
        range of acquisition layouts, and frequencies as well as conductivities
        or resistivities span many orders of magnitude. This makes it hard to
        have a function that fits all purposes. Check the meshes with your
        expert knowledge. Also, the automatic gridding is conservative in its
        estimate, in order to be on the save side (correct results over speed).
        This means, however, that often smaller grids could be used by
        providing the appropriate options in ``gridding_opts``.

    .. note::

        The Simulation-class is currently only implemented for
        ``survey.fixed=False``.


    Parameters
    ----------
    survey : :class:`emg3d.surveys.Survey`
        The survey layout, containing sources, receivers, frequencies, and
        optionally the measured data.

        The survey-data will be modified in place. Provide survey.copy() if you
        want to avoid this.

    grid : :class:`emg3d.meshes.TensorMesh`
        The grid. See :class:`emg3d.meshes.TensorMesh`.

    model : :class:`emg3d.models.Model`
        The model. See :class:`emg3d.models.Model`.

    gridding : str, optional
        Method how the computational grids are computed. Default is 'single'.
        The different methods are:

        - 'same': Same grid as for the input model.
        - 'single': A single grid for all sources and frequencies.
        - 'frequency': Frequency-dependent grids.
        - 'source': Source-dependent grids.
        - 'both': Frequency- and source-dependent grids.
        - 'input': Same as 'single', but instead of automatically generate
          the mesh it has to be provided in ``gridding_opts``.
        - 'dict': Same as 'both', but instead of automatically generate the
          meshes they have to be provided as a ``dict[source][frequency]``
          in ``gridding_opts``.

        See the parameter ``gridding_opts`` for more details.

    gridding_opts : dict or TensorMesh, optional
        Input format depends on ``gridding``:

        - 'same': Nothing, ``gridding_opts`` is not permitted.
        - 'single', 'frequency', 'source', 'both': Described below.
        - 'input': A :class:`emg3d.meshes.TensorMesh`.
        - 'dict': Dictionary of the format ``dict[source][frequency]``
          containing a :class:`emg3d.meshes.TensorMesh` for each
          source-frequency pair.

        The dict in the case of 'single', 'frequency', 'source', 'both' is
        passed to :func:`emg3d.meshes.construct_mesh`; consult the
        corresponding documentation for more information. Parameters that are
        not provided are estimated from the model, grid, and survey using
        :func:`estimate_gridding_opts`, which documentation contains more
        information too.

        There are two notably differences to the parameters described in
        :func:`emg3d.meshes.construct_mesh`:

        - ``vector``: besides the normal possibility it can also be a string
          containing one or several of 'x', 'y', and 'z'. In these cases the
          corresponding dimension of the input mesh is provided as vector.
          See :func:`estimate_gridding_opts`.
        - ``expand``: in the format of ``[property_sea, property_air]``; if
          provided, the input model is expanded up to the seasurface with sea
          water, and an air layer is added. The actual height of the seasurface
          can be defined with the key ``seasurface``. See
          :func:`expand_grid_model`.

    solver_opts : dict, optional
        Passed through to :func:`emg3d.solver.solve`. The dict can contain any
        parameter that is accepted by the :func:`emg3d.solver.solve` except for
        `grid`, `model`, `sfield`, `efield`, `return_info`, and `log`.
        If not provided the following defaults are used:

        - `sslsolver=True`;
        - `semicoarsening=True`;
        - `linerelaxation=True`;
        - `verb=2`;

        Note that these defaults are different from the defaults in
        :func:`emg3d.solver.solve`. The defaults chosen here will be slower in
        many cases, but they are the most robust combination at which you can
        throw most things.

    max_workers : int
        The maximum number of processes that can be used to execute the
        given calls. Default is 4.

    verb : int; optional
        Level of verbosity. Default is 0. These are propagated to the called
        function if not defined above in the possible options.

        - -1: Error.
        - 0: Warning.
        - 1: Info.

    """

    # Gridding descriptions (for repr's).
    _gridding_descr = {
            'same': 'Same grid as for model',
            'single': 'A single grid for all sources and frequencies',
            'frequency': 'Frequency-dependent grids',
            'source': 'Source-dependent grids',
            'both': 'Frequency- and source-dependent grids',
            'input': 'A single, provided grid all sources/frequencies',
            'dict': 'Provided dict of frequency-/source-dependent grids',
            }

    def __init__(self, name, survey, grid, model, max_workers=4,
                 gridding='single', **kwargs):
        """Initiate a new Simulation instance."""

        # Store mandatory inputs (grid and model later).
        self.name = name
        self.survey = survey
        self.max_workers = max_workers
        self.gridding = gridding

        # Get optional inputs with defaults.
        gridding_opts = kwargs.pop('gridding_opts', {}).copy()
        solver_opts = kwargs.pop('solver_opts', {})
        self.verb = kwargs.pop('verb', 0)

        # Store solver options with defaults: The slowest but most robust
        # setting is used, but user-input overwrites defaults if provided.
        # However, verbosity is turned into a log, not real-time verbosity.
        self.solver_opts = {'sslsolver': True, 'semicoarsening': True,
                            'linerelaxation': True, 'verb': 2,
                            **solver_opts, 'return_info': True, 'log': -1}

        # Store original input nCz.
        self._input_nCz = kwargs.pop('_input_nCz', grid.vnC[2])

        # Ensure no kwargs left.
        if kwargs:
            raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}")

        # Fixed surveys are not yet implemented.
        if self.survey.fixed:
            raise NotImplementedError(
                    "Simulation currently only implemented for "
                    "`survey.fixed=False`.")

        # Initiate dictionaries and other values with None's.
        self._dict_grid = self._dict_initiate
        self._dict_model = self._dict_initiate
        self._dict_sfield = self._dict_initiate
        self._dict_efield = self._dict_initiate
        self._dict_hfield = self._dict_initiate
        self._dict_efield_info = self._dict_initiate
        self._gradient = None
        self._misfit = None

        # Check gridding_opts depending on gridding and act upon.
        if self.gridding == 'dict':
            self._dict_grid = gridding_opts
        elif self.gridding == 'input':
            self._grid_single = gridding_opts
        elif self.gridding == 'same':
            if gridding_opts:
                msg = "`gridding_opts` is not permitted if `gridding='same'`"
                raise TypeError(msg)
        else:

            # Expand model by water and air if required.
            expand = gridding_opts.pop('expand', None)
            if expand is not None:
                try:
                    interface = gridding_opts['seasurface']
                except KeyError as e:
                    msg = ("`gridding_opts['seasurface']` is required if "
                           "`gridding_opts['expand']` is provided.")
                    raise KeyError(msg) from e
                grid, model = expand_grid_model(grid, model, expand, interface)

            # Get automatic gridding input.
            # Estimate the parameters from survey and model if not provided.
            self.gridding_opts = estimate_gridding_opts(
                    gridding_opts, grid, model, survey, self._input_nCz)

        # Store grid and model.
        self.grid = grid
        self.model = model

        # Initiate synthetic data with NaN's if they don't exist.
        if 'synthetic' not in self.survey.data.keys():
            self.survey._data['synthetic'] = self.survey.data.observed*np.nan

        # `tqdm`-options; undocumented for the moment.
        # This is likely to change with regards to verbosity and logging.
        self._tqdm_opts = {
                'bar_format': '{desc}: {bar}{n_fmt}/{total_fmt}  [{elapsed}]',
                }

    def __repr__(self):
        return (f"*{self.__class__.__name__}* «{self.name}» "
                f"of {self.survey.__class__.__name__} «{self.survey.name}»\n\n"
                f"- {self.survey.__class__.__name__}: "
                f"{self.survey.shape[0]} sources; "
                f"{self.survey.shape[1]} receivers; "
                f"{self.survey.shape[2]} frequencies\n"
                f"- {self.model.__repr__()}\n"
                f"- Gridding: {self._gridding_descr[self.gridding]}; "
                f"{self._info_grids}")

    def _repr_html_(self):
        return (f"<h3>{self.__class__.__name__} «{self.name}»</h3>"
                f"of {self.survey.__class__.__name__} «{self.survey.name}»<ul>"
                f"<li>{self.survey.__class__.__name__}: "
                f"{self.survey.shape[0]} sources; "
                f"{self.survey.shape[1]} receivers; "
                f"{self.survey.shape[2]} frequencies</li>"
                f"<li>{self.model.__repr__()}</li>"
                f"<li>Gridding: {self._gridding_descr[self.gridding]}; "
                f"{self._info_grids}</li>"
                f"</ul>")

    def copy(self, what='computed'):
        """Return a copy of the Simulation.

        See `to_file` for more information regarding `what`.

        """
        return self.from_dict(self.to_dict(what, True))

    def to_dict(self, what='computed', copy=False):
        """Store the necessary information of the Simulation in a dict.

        See `to_file` for more information regarding `what`.

        """

        if what not in ['computed', 'results', 'all', 'plain']:
            raise TypeError(f"Unrecognized `what`: {what}")

        # If to_dict is called from to_file, it has a _what_to_file attribute.
        if hasattr(self, '_what_to_file'):
            what = self._what_to_file
            delattr(self, '_what_to_file')

        # Initiate dict.
        out = {'name': self.name, '__class__': self.__class__.__name__}

        # Add initiation parameters.
        out['survey'] = self.survey.to_dict()
        out['grid'] = self.grid.to_dict()
        out['model'] = self.model.to_dict()
        out['max_workers'] = self.max_workers
        out['gridding'] = self.gridding
        out['solver_opts'] = self.solver_opts

        # Clean unwanted data if plain.
        if what == 'plain':
            for key in ['synthetic', 'residual', 'weights']:
                if key in out['survey']['data'].keys():
                    del out['survey']['data'][key]

        # Put provided grids back on gridding.
        if self.gridding == 'input':
            out['gridding_opts'] = self._grid_single
        elif self.gridding == 'dict':
            out['gridding_opts'] = self._dict_grid
        elif self.gridding != 'same':

            gopts = self.gridding_opts

            # Take care of map.
            if 'mapping' in gopts.keys():
                if not isinstance(gopts['mapping'], str):
                    gopts['mapping'] = gopts['mapping'].name

            # Take care of tuples and lists for h5/npz.
            # Ideally, this should be dealt with in emg3d.io.
            for key in ['domain', 'vector', 'stretching', 'min_width_limits']:
                if (key in gopts.keys() and
                        isinstance(gopts[key], (list, tuple))):
                    this_type = type(gopts[key]).__name__
                    gopts[key] = {i: v for i, v in enumerate(gopts[key])}
                    gopts[key]['__type__'] = this_type

            out['gridding_opts'] = gopts

        out['_input_nCz'] = self._input_nCz

        # Store wanted dicts.
        if what in ['computed', 'all']:
            for name in ['_dict_efield', '_dict_efield_info', '_dict_hfield',
                         '_dict_bfield', '_dict_bfield_info']:
                if hasattr(self, name):
                    out[name] = getattr(self, name)

            if what == 'all':
                for name in ['_dict_grid', '_dict_model', '_dict_sfield']:
                    if hasattr(self, name):
                        out[name] = getattr(self, name)

        # Store gradient and misfit.
        if what in ['computed', 'results', 'all']:
            out['gradient'] = self._gradient
            out['misfit'] = self._misfit

        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert dictionary into :class:`Simulation` instance.


        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`Simulation.to_dict`.

        Returns
        -------
        obj : :class:`Simulation` instance

        """
        from emg3d import io

        try:

            # gridding options, backwards compatible.
            gridding = inp['gridding']
            gridding_opts = inp.get('gridding_opts', {})

            # De-serialize lists and tuples.
            # Ideally, this should be dealt with in emg3d.io.
            if gridding not in ['input', 'dict', 'same']:
                props = ['domain', 'vector', 'stretching', 'min_width_limits']
                for key in props:
                    if key in gridding_opts.keys():
                        if '__type__' in gridding_opts[key].keys():
                            this_type = gridding_opts[key].pop('__type__')
                            out = [d for d in gridding_opts[key].values()]
                            if this_type == 'tuple':
                                out = tuple(out)
                            gridding_opts[key] = out

            # Initiate class.
            out = cls(
                    name=inp['name'],
                    survey=surveys.Survey.from_dict(inp['survey']),
                    grid=meshes.TensorMesh.from_dict(inp['grid']),
                    model=models.Model.from_dict(inp['model']),
                    max_workers=inp['max_workers'],
                    gridding=gridding,
                    solver_opts=inp['solver_opts'],
                    gridding_opts=gridding_opts,
                    _input_nCz=inp.get('_input_nCz', len(inp['grid']['hz']))
                    )

            # Add existing derived/computed properties.
            data = ['_dict_grid', '_dict_model', '_dict_sfield',
                    '_dict_hfield', '_dict_efield', '_dict_efield_info',
                    '_dict_bfield', '_dict_bfield_info']
            for name in data:
                if name in inp.keys():
                    values = inp.get(name)

                    # Storing to_file makes strings out of the freq-keys.
                    # Undo this.
                    new_values = {}
                    for src, val in values.items():
                        new_values[src] = {}
                        for freq, v in val.items():
                            new_values[src][float(freq)] = val.get(freq)

                    # De-serialize Model, Field, and TensorMesh instances.
                    io._dict_deserialize(new_values)

                    setattr(out, name, new_values)

            data = ['gradient', 'misfit']
            for name in data:
                if name in inp.keys():
                    setattr(out, '_'+name, inp.get(name))

            # For backwards compatibility < v0.16.0; remove eventually.
            if 'data' in inp.keys():
                warnings.warn("Simulation-dict is outdated; store with new "
                              "version of `emg3d`.", FutureWarning)
                for name in inp['data'].keys():
                    out.data[name] = out.data.observed*np.nan
                    out.data[name][...] = inp['data'][name]

            return out

        except KeyError as e:
            raise KeyError(f"Variable {e} missing in `inp`.") from e

    def to_file(self, fname, what='computed', name='simulation', **kwargs):
        """Store Simulation to a file.

        Parameters
        ----------
        fname : str
            File name inclusive ending, which defines the used data format.
            Implemented are currently:

            - `.h5` (default): Uses `h5py` to store inputs to a hierarchical,
              compressed binary hdf5 file. Recommended file format, but
              requires the module `h5py`. Default format if ending is not
              provided or not recognized.
            - `.npz`: Uses `numpy` to store inputs to a flat, compressed binary
              file. Default format if `h5py` is not installed.
            - `.json`: Uses `json` to store inputs to a hierarchical, plain
              text file.

        what : str
            What to store. Currently implemented:

            - 'computed' (default):
              Stores all computed properties: electric fields and responses at
              receiver locations.
            - 'results':
              Stores only the response at receiver locations.
            - 'all':
              Stores everything.
            - 'plain':
              Only stores the plain Simulation (as initiated).

        name : str
            Name under which the survey is stored within the file.

        kwargs : Keyword arguments, optional
            Passed through to :func:`emg3d.io.save`.

        """
        from emg3d import io

        # Add what to self, will be removed in to_dict.
        self._what_to_file = what

        kwargs[name] = self                # Add simulation to dict.
        kwargs['collect_classes'] = False  # Ensure classes are not collected.
        # If verb is not defined, use verbosity of simulation.
        if 'verb' not in kwargs:
            kwargs['verb'] = self.verb

        return io.save(fname, **kwargs)

    @classmethod
    def from_file(cls, fname, name='simulation', **kwargs):
        """Load Simulation from a file.

        Parameters
        ----------
        fname : str
            File name including extension. Used backend depends on the file
            extensions:

            - '.npz': numpy-binary
            - '.h5': h5py-binary (needs `h5py`)
            - '.json': json

        name : str
            Name under which the simulation is stored within the file.

        kwargs : Keyword arguments, optional
            Passed through to :func:`emg3d.io.load`.

        Returns
        -------
        simulation : :class:`Simulation`
            The simulation that was stored in the file.

        """
        from emg3d import io
        out = io.load(fname, **kwargs)
        if 'verb' in kwargs and kwargs['verb'] < 0:
            return out[0][name], out[1]
        else:
            return out[name]

    # GET FUNCTIONS
    def get_grid(self, source, frequency):
        """Return computational grid of the given source and frequency."""
        freq = float(frequency)

        # Return grid if it exists already.
        if self._dict_grid[source][freq] is not None:
            return self._dict_grid[source][freq]

        # Act depending on gridding:
        if self.gridding == 'same':  # Same grid as for provided model.

            # Store link to grid.
            self._dict_grid[source][freq] = self.grid

        elif self.gridding == 'frequency':  # Frequency-dependent grids.

            # Initiate dict.
            if not hasattr(self, '_grid_frequency'):
                self._grid_frequency = {}

            # Get grid for this frequency if not yet computed.
            if freq not in self._grid_frequency.keys():

                # Get grid and store it.
                inp = {**self.gridding_opts, 'frequency': freq}
                self._grid_frequency[freq] = meshes.construct_mesh(**inp)

            # Store link to grid.
            self._dict_grid[source][freq] = self._grid_frequency[freq]

        elif self.gridding == 'source':  # Source-dependent grids.

            # Initiate dict.
            if not hasattr(self, '_grid_source'):
                self._grid_source = {}

            # Get grid for this source if not yet computed.
            if source not in self._grid_source.keys():

                # Get grid and store it.
                center = self.survey.sources[source].coordinates[:3]
                inp = {**self.gridding_opts, 'center': center}
                self._grid_source[source] = meshes.construct_mesh(**inp)

            # Store link to grid.
            self._dict_grid[source][freq] = self._grid_source[source]

        elif self.gridding == 'both':  # Src- & freq-dependent grids.

            # Get grid and store it.
            center = self.survey.sources[source].coordinates[:3]
            inp = {**self.gridding_opts, 'frequency': freq, 'center': center}
            self._dict_grid[source][freq] = meshes.construct_mesh(**inp)

        else:  # Use a single grid for all sources and receivers.
            # Default case; catches 'single' but also anything else.

            # Get grid if not yet computed.
            if not hasattr(self, '_grid_single'):

                # Get grid and store it.
                self._grid_single = meshes.construct_mesh(**self.gridding_opts)

            # Store link to grid.
            self._dict_grid[source][freq] = self._grid_single

        # Use recursion to return grid.
        return self.get_grid(source, frequency)

    def get_model(self, source, frequency):
        """Return model on the grid of the given source and frequency."""
        freq = float(frequency)

        # Return model if it exists already.
        if self._dict_model[source][freq] is not None:
            return self._dict_model[source][freq]

        # Act depending on gridding:
        if self.gridding == 'same':  # Same grid as for provided model.

            # Store link to model.
            self._dict_model[source][freq] = self.model

        elif self.gridding == 'frequency':  # Frequency-dependent grids.

            # Initiate dict.
            if not hasattr(self, '_model_frequency'):
                self._model_frequency = {}

            # Get model for this frequency if not yet computed.
            if freq not in self._model_frequency.keys():
                self._model_frequency[freq] = self.model.interpolate2grid(
                        self.grid, self.get_grid(source, freq))

            # Store link to model.
            self._dict_model[source][freq] = self._model_frequency[freq]

        elif self.gridding == 'source':  # Source-dependent grids.

            # Initiate dict.
            if not hasattr(self, '_model_source'):
                self._model_source = {}

            # Get model for this source if not yet computed.
            if source not in self._model_source.keys():
                self._model_source[source] = self.model.interpolate2grid(
                        self.grid, self.get_grid(source, freq))

            # Store link to model.
            self._dict_model[source][freq] = self._model_source[source]

        elif self.gridding == 'both':  # Src- & freq-dependent grids.

            # Get model and store it.
            self._dict_model[source][freq] = self.model.interpolate2grid(
                        self.grid, self.get_grid(source, freq))

        else:  # Use a single grid for all sources and receivers.
            # Default case; catches 'single' but also anything else.

            # Get model if not yet computed.
            if not hasattr(self, '_model_single'):
                self._model_single = self.model.interpolate2grid(
                        self.grid, self.get_grid(source, freq))

            # Store link to model.
            self._dict_model[source][freq] = self._model_single

        # Use recursion to return model.
        return self.get_model(source, frequency)

    def get_sfield(self, source, frequency):
        """Return source field for given source and frequency."""
        freq = float(frequency)

        # Get source field if it is not stored yet.
        if self._dict_sfield[source][freq] is None:

            # Get source and source strength.
            src = self.survey.sources[source]
            if hasattr(src, 'strength'):
                strength = src.strength
            else:
                strength = 0

            sfield = fields.get_source_field(
                    grid=self.get_grid(source, frequency),
                    src=src.coordinates,
                    freq=frequency,
                    strength=strength,
                    electric=src.electric)

            self._dict_sfield[source][freq] = sfield

        # Return source field.
        return self._dict_sfield[source][freq]

    def get_efield(self, source, frequency, **kwargs):
        """Return electric field for given source and frequency."""
        freq = float(frequency)

        # Get call_from_compute and ensure no kwargs are left.
        call_from_compute = kwargs.pop('call_from_compute', False)
        call_from_hfield = kwargs.pop('call_from_hfield', False)
        if kwargs:
            raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}")

        # Compute electric field if it is not stored yet.
        if self._dict_efield[source][freq] is None:

            # Input parameters.
            solver_input = {
                **self.solver_opts,
                'grid': self.get_grid(source, freq),
                'model': self.get_model(source, freq),
                'sfield': self.get_sfield(source, freq),
            }

            # Compute electric field.
            efield, info = solver.solve(**solver_input)

            # Store electric field and info.
            self._dict_efield[source][freq] = efield
            self._dict_efield_info[source][freq] = info

            if not call_from_hfield:

                # Clean corresponding hfield, so it will be recomputed.
                del self._dict_hfield[source][freq]
                self._dict_hfield[source][freq] = None

                # Store electric and magnetic responses at receiver locations.
                self._store_responses(source, frequency)

        # Return electric field.
        if call_from_compute:
            return (self._dict_efield[source][freq],
                    self._dict_efield_info[source][freq],
                    self._dict_hfield[source][freq],
                    self.data.synthetic.loc[source, :, freq].data)
        else:
            return self._dict_efield[source][freq]

    def get_hfield(self, source, frequency, **kwargs):
        """Return magnetic field for given source and frequency."""
        freq = float(frequency)

        # If magnetic field not computed yet compute it.
        if self._dict_hfield[source][freq] is None:

            self._dict_hfield[source][freq] = fields.get_h_field(
                    self.get_grid(source, freq),
                    self.get_model(source, freq),
                    self.get_efield(source, freq,
                                    call_from_hfield=True, **kwargs))

            # Store electric and magnetic responses at receiver locations.
            self._store_responses(source, frequency)

        # Return magnetic field.
        return self._dict_hfield[source][freq]

    def _store_responses(self, source, frequency):
        """Return electric and magnetic fields at receiver locations."""
        freq = float(frequency)

        # Get receiver coordinates.
        rec_coords = self.survey.rec_coords
        rec_types = self.survey.rec_types

        # For fixed surveys:
        # rec_coords = self.survey.rec_coords[source]

        # Store electric receivers.
        if rec_types.count(True):

            # Extract data at receivers.
            erec = np.nonzero(rec_types)[0]
            resp = fields.get_receiver_response(
                    grid=self.get_grid(source, freq),
                    field=self.get_efield(source, freq),
                    rec=tuple(np.array(rec_coords)[:, erec])
            )

            # Store the receiver response.
            self.data.synthetic.loc[source, :, freq][erec] = resp

        # Store magnetic receivers.
        if rec_types.count(False):

            # Extract data at receivers.
            mrec = np.nonzero(np.logical_not(rec_types))[0]
            resp = fields.get_receiver_response(
                    grid=self.get_grid(source, freq),
                    field=self.get_hfield(source, freq),
                    rec=tuple(np.array(rec_coords)[:, mrec])
            )

            # Store the receiver response.
            self.data.synthetic.loc[source, :, freq][mrec] = resp

    def get_efield_info(self, source, frequency):
        """Return the solver information of the corresponding computation."""
        return self._dict_efield_info[source][float(frequency)]

    # ASYNCHRONOUS COMPUTATION
    def _get_efield(self, inp):
        """Wrapper of `get_efield` for `concurrent.futures`."""
        return self.get_efield(*inp, call_from_compute=True)

    def compute(self, observed=False, **kwargs):
        """Compute efields asynchronously for all sources and frequencies.

        Parameters
        ----------
        observed : bool
            If True, it stores the current result also as observed model.
            This is usually done for pure forward modelling (not inversion).
            It will as such be stored within the survey. If the survey has
            either `relative_error` or `noise_floor`, random Gaussian noise
            with std will be added to the `data.observed` (not to
            data.synthetic). Also, data below the noise floor will be set to
            NaN.

        min_offset : float
            Default is 0.0. Data in `data.observed` where the offset <
            min_offset are set to NaN.

        """
        srcfreq = self._srcfreq.copy()

        # We remove the ones that were already computed.
        remove = []
        for src, freq in srcfreq:
            if self._dict_efield[src][freq] is not None:
                remove += [(src, freq)]
        for src, freq in remove:
            srcfreq.remove((src, freq))

        # Ensure grids, models, and source fields are computed.
        #
        # => This could be done within the field computation. But then it might
        #    have to be done multiple times even if 'single' or 'same' grid.
        #    Something to keep in mind.
        #    For `gridding='same'` it does not really matter.
        for src, freq in srcfreq:
            _ = self.get_grid(src, freq)
            _ = self.get_model(src, freq)
            _ = self.get_sfield(src, freq)

        # Initiate futures-dict to store output.
        out = process_map(
                self._get_efield,
                srcfreq,
                max_workers=self.max_workers,
                **{'desc': 'Compute efields', **self._tqdm_opts},
        )

        # Loop over src-freq combinations to extract and store.
        for i, (src, freq) in enumerate(srcfreq):

            # Store efield and solver info.
            self._dict_efield[src][freq] = out[i][0]
            self._dict_efield_info[src][freq] = out[i][1]
            self._dict_hfield[src][freq] = out[i][2]

            # Store responses at receivers.
            self.data['synthetic'].loc[src, :, freq] = out[i][3]

        # Print solver info.
        self.print_solver_info('efield', verb=self.verb)

        # If it shall be used as observed data save a copy.
        if observed:

            self.data['observed'] = self.data['synthetic'].copy()

            # Add noise noise_floor and/or relative_error given.
            if self.survey.standard_deviation is not None:

                # Create noise.
                std = self.survey.standard_deviation
                random = np.random.randn(self.survey.size*2)
                noise_re = std*random[::2].reshape(self.survey.shape)
                noise_im = std*random[1::2].reshape(self.survey.shape)

                # Add noise to observed data.
                self.data['observed'].data += noise_re + 1j*noise_im

            # Set data below the noise floor to NaN.
            if self.data.noise_floor is not None:
                min_amp = abs(self.data.synthetic.data) < self.data.noise_floor
                self.data['observed'].data[min_amp] = np.nan + 1j*np.nan

            # Set near-offsets to NaN.
            offsets = sl.norm(
                np.array(self.survey.rec_coords[:3])[:, None, :] -
                np.array(self.survey.src_coords[:3])[:, :, None],
                axis=0,
                check_finite=False,
            )
            min_off = offsets < kwargs.get('min_offset', 0.0)
            self.data['observed'].data[min_off] = np.nan + 1j*np.nan

    # DATA
    @property
    def data(self):
        """Shortcut to survey.data."""
        return self.survey.data

    # OPTIMIZATION
    @property
    def gradient(self):
        """Return the gradient of the misfit function.

        See :func:`emg3d.optimize.gradient`.

        """
        # Compute it if not stored already.
        if self._gradient is None:
            self._gradient = optimize.gradient(self)

        return self._gradient[:, :, :self._input_nCz]

    @property
    def misfit(self):
        """Return the misfit function.

        See :func:`emg3d.optimize.misfit`.

        """
        # Compute it if not stored already.
        if self._misfit is None:
            self._misfit = optimize.misfit(self)

        return self._misfit

    # UTILS
    def clean(self, what='computed'):
        """Clean part of the data base.

        Parameters
        ----------
        what : str
            What to clean. Currently implemented:

            - 'computed' (default):
              Removes all computed properties: electric and magnetic fields and
              responses at receiver locations.
            - 'keepresults':
              Removes everything  except for the responses at receiver
              locations.
            - 'all':
              Removes everything (leaves it plain as initiated).

        """

        if what not in ['computed', 'keepresults', 'all']:
            raise TypeError(f"Unrecognized `what`: {what}")

        # Clean data/model/sfield-dicts.
        if what in ['keepresults', 'all']:
            for name in ['_dict_grid', '_dict_model', '_dict_sfield']:
                delattr(self, name)
                setattr(self, name, self._dict_initiate)

        # Clean field-dicts.
        if what in ['computed', 'keepresults', 'all']:

            # These exist always and have to be initiated.
            for name in ['_dict_efield', '_dict_efield_info', '_dict_hfield']:
                delattr(self, name)
                setattr(self, name, self._dict_initiate)

            # These only exist with gradient; don't initiate them.
            for name in ['_dict_bfield', '_dict_bfield_info']:
                if hasattr(self, name):
                    delattr(self, name)

        # Clean data.
        if what in ['computed', 'all']:
            for key in ['residual', 'weight']:
                if key in self.data.keys():
                    del self.data[key]
            self.data['synthetic'] = self.data.observed*np.nan
            for name in ['_gradient', '_misfit']:
                delattr(self, name)
                setattr(self, name, None)

    @property
    def _dict_initiate(self):
        """Returns a dict of the structure `dict[source][freq]=None`."""
        return {src: {freq: None for freq in self.survey.frequencies}
                for src in self.survey.sources.keys()}

    @property
    def _srcfreq(self):
        """Return list of all source-frequency pairs."""

        if getattr(self, '__srcfreq', None) is None:
            self.__srcfreq = list(
                    itertools.product(self.survey.sources.keys(),
                                      self.survey.frequencies))

        return self.__srcfreq

    @property
    def _info_grids(self):
        """Return a string with "min {- max}" grid size."""
        if self.gridding == 'same':
            min_nC = self.grid.nC
            min_vC = self.grid.vnC
            has_minmax = False
        elif self.gridding in ['single', 'input']:
            grid = self.get_grid(self._srcfreq[0][0], self._srcfreq[0][1])
            min_nC = grid.nC
            min_vC = grid.vnC
            has_minmax = False
        else:
            min_nC = 1e100
            max_nC = 0
            for src, freq in self._srcfreq:
                grid = self.get_grid(src, freq)
                if grid.nC > max_nC:
                    max_nC = grid.nC
                    max_vC = grid.vnC
                if grid.nC < min_nC:
                    min_nC = grid.nC
                    min_vC = grid.vnC
            has_minmax = min_nC != max_nC
        info = f"{min_vC[0]} x {min_vC[1]} x {min_vC[2]} ({min_nC:,})"
        if has_minmax:
            info += f" - {max_vC[0]} x {max_vC[1]} x {max_vC[2]} ({max_nC:,})"
        return info

    def print_grid_info(self, verb=1, return_info=False):
        """Print info for all generated grids."""

        def get_grid_info(src, freq):
            """Return grid info for given source and frequency."""
            grid = self.get_grid(src, freq)
            out = ''
            if verb != 0 and hasattr(grid, 'construct_mesh_info'):
                out += grid.construct_mesh_info
            out += grid.__repr__()
            return out

        # Act depending on gridding:
        out = ""
        if self.gridding == 'frequency':

            # Loop over frequencies.
            for freq in self.survey.frequencies:
                out += f"= Source: all; Frequency: {freq} Hz =\n"
                out += get_grid_info(self._srcfreq[0][0], freq)

        elif self.gridding == 'source':

            # Loop over sources.
            for src in self.survey.sources.keys():
                out += f"= Source: {src}; Frequency: all =\n"
                out += get_grid_info(src, self._srcfreq[0][1])

        elif self.gridding == 'both':

            # Loop over sources, frequencies.
            for src, freq in self._srcfreq:
                out += f"= Source: {src}; Frequency: {freq} Hz =\n"
                out += get_grid_info(src, freq)

        else:  # same, input, single

            out += "= Source: all; Frequency: all =\n"
            out += get_grid_info(self._srcfreq[0][0], self._srcfreq[0][1])

        if return_info:
            return out
        elif out:
            print(out)

    def print_solver_info(self, field='efield', verb=1, return_info=False):
        """Print solver info."""

        # Get info dict.
        info = getattr(self, f"_dict_{field}_info", {})
        out = ""

        # Loop over sources and frequencies.
        if verb > -1:
            for src, freq in self._srcfreq:
                cinfo = info[src][freq]

                # Print if verbose or not converged.
                if cinfo is not None and (verb > 0 or cinfo['exit'] != 0):

                    # Initial message.
                    if not out:
                        out += "\n"
                        if verb > 0:
                            out += f"    - SOLVER INFO <{field}> -\n\n"

                    # Source and frequency info.
                    out += f"= Source {src}; Frequency {freq} Hz ="

                    # Print log depending on solver and simulation verbosities.
                    if verb == 0 or self.solver_opts['verb'] not in [1, 2]:
                        out += f" {cinfo['exit_message']}\n"

                    if verb > 0 and self.solver_opts['verb'] > 2:
                        out += f"\n{cinfo['log']}\n"

                    if verb > 0 and self.solver_opts['verb'] in [1, 2]:
                        out += f" {cinfo['log'][12:]}"

        if return_info:
            return out
        elif out:
            print(out)

    # BACKWARDS PROPAGATING FIELD
    def _get_bfields(self, inp):
        """Return back-propagated electric field for given inp (src, freq)."""

        # Input parameters.
        solver_input = {
            **self.solver_opts,
            'grid': self.get_grid(*inp),
            'model': self.get_model(*inp),
            'sfield': self._get_rfield(*inp),  # Residual field.
        }

        # Compute and return back-propagated electric field.
        return solver.solve(**solver_input)

    def _bcompute(self):
        """Compute bfields asynchronously for all sources and frequencies."""

        # Initiate futures-dict to store output.
        out = process_map(
                self._get_bfields,
                self._srcfreq,
                max_workers=self.max_workers,
                **{'desc': 'Back-propagate ', **self._tqdm_opts},
        )

        # Store back-propagated electric field and info.
        if not hasattr(self, '_dict_bfield'):
            self._dict_bfield = self._dict_initiate
            self._dict_bfield_info = self._dict_initiate

        # Loop over src-freq combinations to extract and store.
        for i, (src, freq) in enumerate(self._srcfreq):

            # Store bfield and solver info.
            self._dict_bfield[src][freq] = out[i][0]
            self._dict_bfield_info[src][freq] = out[i][1]

        # Print solver info.
        self.print_solver_info('bfield', verb=self.verb)

    def _get_rfield(self, source, frequency):
        """Return residual source field for given source and frequency."""

        freq = float(frequency)
        grid = self.get_grid(source, frequency)

        # Initiate empty field
        ResidualField = fields.SourceField(grid, freq=frequency)

        # Loop over receivers, input as source.
        for name, rec in self.survey.receivers.items():

            # Strength: in get_source_field the strength is multiplied with
            # iwmu; so we undo this here.
            residual = self.data.residual.loc[source, name, freq].data
            if np.isnan(residual):
                continue
            strength = residual.conj()
            strength *= self.data.weights.loc[source, name, freq].data.conj()
            strength /= ResidualField.smu0

            # Magnetic receivers: the adjoint gradient is formulated for the
            # electric fields with ``S`` in Equation (1) of [PlMu08]_. For
            # magnetic receivers we take M instead, which does not have the
            # factor iwmu; we scale here the source strength accordingly.
            # Or, because of loop (B not H).
            if not rec.electric:
                strength /= ResidualField.smu0

            # If strength is zero (very unlikely), get_source_field would
            # return a normalized field for a unit source. However, in this
            # case we do not want that.
            if strength != 0:
                ResidualField += fields.get_source_field(
                    grid=grid,
                    src=rec.coordinates,
                    freq=frequency,
                    strength=strength,
                    electric=rec.electric,
                )

        return ResidualField


# HELPER FUNCTIONS
def expand_grid_model(grid, model, expand, interface):
    """Expand model and grid according to provided parameters.

    Expand the grid and corresponding model in positive z-direction from the
    end of the grid to the interface with ``expand[0]``, and a 100 m thick
    layer above the interface with ``expand[1]``.

    The provided properties are taken as isotropic; ``mu_r`` and ``epsilon_r``
    are expanded with ones, if necessary.

    The ``interface`` is usually the sea-surface, and ``expand`` is therefore
    ``[property_sea, property_air]``.

    Parameters
    ----------
    grid : :class:`emg3d.meshes.TensorMesh`
        The grid.

    model : :class:`emg3d.models.Model`
        The model.

    expand : list
        The two properties below and above the interface:
        ``[below_interface, above_interface]``.

    interface : float
        Interface between the two properties in ``expand``.


    Returns
    -------
    grid : :class:`emg3d.meshes.TensorMesh`
        Expanded grid.

    model : :class:`emg3d.models.Model`
        Expanded model.

    """

    def extend_property(prop, add_values, nadd):
        """Expand property `model.prop`, IF it is not None."""

        if getattr(model, '_'+prop) is None:
            prop_ext = None

        else:
            prop_ext = np.zeros((grid.vnC[0], grid.vnC[1], grid.vnC[2]+nadd))
            prop_ext[:, :, :-nadd] = getattr(model, prop)
            if nadd == 2:
                prop_ext[:, :, -2] = add_values[0]
            prop_ext[:, :, -1] = add_values[1]

        return prop_ext

    # Initiate.
    nzadd = 0
    hz_ext = grid.h[2]

    # Fill-up property_below.
    if grid.nodes_z[-1] < interface-0.05:  # At least 5 cm.
        hz_ext = np.r_[hz_ext, interface-grid.nodes_z[-1]]
        nzadd += 1

    # Add 100 m of property_above.
    if grid.nodes_z[-1] <= interface+0.001:  # +1mm
        hz_ext = np.r_[hz_ext, 100]
        nzadd += 1

    if nzadd > 0:
        # Extend properties.
        property_x = extend_property('property_x', expand, nzadd)
        property_y = extend_property('property_y', expand, nzadd)
        property_z = extend_property('property_z', expand, nzadd)
        mu_r = extend_property('mu_r', [1, 1], nzadd)
        epsilon_r = extend_property('epsilon_r', [1, 1], nzadd)

        # Create extended grid and model.
        grid = meshes.TensorMesh(
                [grid.h[0], grid.h[1], hz_ext], origin=grid.origin)
        model = models.Model(
                grid, property_x, property_y, property_z, mu_r,
                epsilon_r, mapping=model.map.name)

    return grid, model


def estimate_gridding_opts(gridding_opts, grid, model, survey, input_nCz=None):
    """Estimate parameters for automatic gridding.

    Automatically determines the required gridding options from the provided
    grid, model, and survey, if they are not provided in ``gridding_opts``.

    The dict ``gridding_opts`` can contain any input parameter taken by
    :func:`emg3d.meshes.construct_mesh`, see the corresponding documentation
    for more details with regards to the possibilities.

    Different keys of ``gridding_opts`` are treated differently:

    - The following parameters are estimated from the ``model`` if not
      provided:

      - ``properties``: lowest conductivity / highest resistivity in the
        outermost layer in a given direction. This is usually air in x/y and
        positive z. Note: This is very conservative. If you go into deeper
        water you could provide less conservative values.
      - ``mapping``: taken from model.

    - The following parameters are estimated from the ``survey`` if not
      provided:

      - ``frequency``: average (on log10-scale) of all frequencies.
      - ``center``: center of all sources.
      - ``domain``: from ``vector``, if provided, or

        - in x/y-directions: extent of sources and receivers plus 10% on each
          side, ensuring ratio of 3.
        - in z-direction: extent of sources and receivers, ensuring ratio of 2
          to horizontal dimension; 1/10 tenth up, 9/10 down.

        The ratio means that it is enforced that the survey dimension in x or
        y-direction is not smaller than a third of the survey dimension in the
        other direction. If not, the smaller dimension is expanded
        symmetrically. Similarly in the vertical direction, which must be at
        least half the dimension of the maximum horizontal dimension or 5 km,
        whatever is smaller. Otherwise it is expanded in a ratio of 9 parts
        downwards, one part upwards.

    - The following parameter is taken from the ``grid`` if provided as a
      string:

      - ``vector``: This is the only real "difference" to the inputs of
        :func:`emg3d.meshes.construct_mesh`. The normal input is accepted, but
        it can also be a string contain any combination of 'x', 'y', and 'z'.
        All directions contained in this string are then taken from the
        provided grid. E.g., if ``gridding_opts['vector']='xz'`` it will take
        the x- and z-directed vectors from the grid.

    - The following parameters are simply passed along if they are provided,
      nothing is done otherwise:

      - ``vector``
      - ``distance``
      - ``stretching``
      - ``seasurface``
      - ``cell_numbers``
      - ``lambda_factor``
      - ``lambda_from_center``
      - ``max_buffer``
      - ``min_width_limits``
      - ``min_width_pps``
      - ``verb``


    Parameters
    ----------
    gridding_opts : dict
        Containing input parameters to provide to
        :func:`emg3d.meshes.construct_mesh`. See the corresponding
        documentation and the explanations above.

    grid : :class:`emg3d.meshes.TensorMesh`
        The grid.

    model : :class:`emg3d.models.Model`
        The model.

    survey : :class:`emg3d.surveys.Survey`
        The survey.

    input_nCz : int, optional
        If :func:`expand_grid_model` was used, `input_nCz` corresponds to the
        original ``grid.vnC[2]``.


    Returns
    -------
    gridding_opts : dict
        Dict to provide to :func:`emg3d.meshes.construct_mesh`.

    """
    # Initiate new gridding_opts.
    gopts = {}

    # Optional values that we only include if provided.
    for name in ['stretching', 'seasurface', 'cell_numbers', 'lambda_factor',
                 'lambda_from_center', 'max_buffer', 'min_width_limits',
                 'min_width_pps', 'verb']:
        if name in gridding_opts.keys():
            gopts[name] = gridding_opts.pop(name)

    # Mapping defaults to model map.
    gopts['mapping'] = gridding_opts.pop('mapping', model.map)

    # Frequency defaults to average frequency (log10).
    freq = 10**np.mean(np.log10(survey.frequencies))
    gopts['frequency'] = gridding_opts.pop('frequency', freq)

    # Center defaults to center of all sources.
    center = tuple([np.mean(survey.src_coords[i]) for i in range(3)])
    gopts['center'] = gridding_opts.pop('center', center)

    # Vector.
    vector = gridding_opts.pop('vector', None)
    if isinstance(vector, str):
        # If vector is a string we take the corresponding vectors from grid.
        vector = (
                grid.nodes_x if 'x' in vector.lower() else None,
                grid.nodes_y if 'y' in vector.lower() else None,
                grid.nodes_z[:input_nCz] if 'z' in vector.lower() else None,
        )
        gopts['vector'] = vector

    elif vector is not None:
        # In this case vector was provided, and we include it like this.
        gopts['vector'] = vector

    # Distance.
    distance = gridding_opts.pop('distance', None)
    if distance is not None:
        gopts['distance'] = distance

    # Properties defaults to lowest conductivities (AFTER model expansion).
    properties = gridding_opts.pop('properties', None)
    if properties is None:

        # Get map (in principle the map in gridding_opts could be different
        # from the map in the model).
        m = gopts['mapping']
        if isinstance(m, str):
            m = getattr(maps, 'Map'+m)()

        # Minimum conductivity of all values (x, y, z).
        def get_min(ix, iy, iz):
            """Get minimum: very conservative/costly, but avoiding problems."""

            # Collect all x (y, z) values.
            data = np.array([])
            for p in ['x', 'y', 'z']:
                prop = getattr(model, 'property_'+p)
                if getattr(model, '_property_'+p) is None:
                    continue
                elif prop.ndim == 1:
                    data = np.r_[data, model.map.backward(prop)]
                else:
                    prop = model.map.backward(prop[ix, iy, iz])
                    data = np.r_[data, np.min(prop)]

            # Return minimum conductivity (on mapping).
            return m.forward(min(data))

        # Buffer properties.
        xneg = get_min(0, slice(None), slice(None))
        xpos = get_min(-1, slice(None), slice(None))
        yneg = get_min(slice(None), 0, slice(None))
        ypos = get_min(slice(None), -1, slice(None))
        zneg = get_min(slice(None), slice(None), 0)
        zpos = get_min(slice(None), slice(None), -1)

        # Source property.
        ix = np.argmin(abs(grid.nodes_x - gopts['center'][0]))
        iy = np.argmin(abs(grid.nodes_y - gopts['center'][1]))
        iz = np.argmin(abs(grid.nodes_z - gopts['center'][2]))
        source = get_min(ix, iy, iz)

        properties = [source, xneg, xpos, yneg, ypos, zneg, zpos]

    gopts['properties'] = properties

    # Domain; default taken from survey.
    domain = gridding_opts.pop('domain', None)

    def get_dim_diff(i):
        """Return ([min, max], dim) of inp.

        Take it from domain if provided, else from vector if provided, else
        from survey, adding 10% on each side).
        """
        if domain is not None and domain[i] is not None:
            # domain is provided.
            dim = domain[i]
            diff = np.diff(dim)[0]
            get_it = False

        elif vector is not None and vector[i] is not None:
            # vector is provided.
            dim = [np.min(vector[i]), np.max(vector[i])]
            diff = np.diff(dim)[0]
            get_it = False

        elif distance is not None and distance[i] is not None:
            # distance is provided.
            dim = None
            diff = abs(distance[i][0]) + abs(distance[i][1])
            get_it = False

        else:
            # Get it from survey, add 5 % on each side.
            inp = np.r_[survey.src_coords[i], survey.rec_coords[i]]
            dim = [min(inp), max(inp)]
            diff = np.diff(dim)[0]
            dim = [min(inp)-diff/10, max(inp)+diff/10]
            diff = np.diff(dim)[0]
            get_it = True

        return dim, diff, get_it

    xdim, xdiff, get_x = get_dim_diff(0)
    ydim, ydiff, get_y = get_dim_diff(1)
    zdim, zdiff, get_z = get_dim_diff(2)

    # Ensure the ratio xdim:ydim is at most 3.
    if get_y and xdiff/ydiff > 3:
        diff = round((xdiff/3.0 - ydiff)/2.0)
        ydim = [ydim[0]-diff, ydim[1]+diff]
    elif get_x and ydiff/xdiff > 3:
        diff = round((ydiff/3.0 - xdiff)/2.0)
        xdim = [xdim[0]-diff, xdim[1]+diff]

    # Ensure the ratio zdim:horizontal is at most 2.
    hdist = min(10000, max(xdiff, ydiff))
    if get_z and hdist/zdiff > 2:
        diff = round((hdist/2.0 - zdiff)/10.0)
        zdim = [zdim[0]-9*diff, zdim[1]+diff]

    # Collect
    domain = (xdim, ydim, zdim)

    gopts['domain'] = domain

    # Ensure no gridding_opts left.
    if gridding_opts:
        print(gridding_opts)
        raise TypeError(
                f"Unexpected gridding_opts: {list(gridding_opts.keys())}")

    # Return gridding_opts.
    return gopts



"""
A survey stores a set of sources, receivers, and the measured data.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import warnings
from copy import deepcopy
from dataclasses import dataclass

import numpy as np

try:
    import xarray
except ImportError:
    xarray = None

from emg3d import utils

__all__ = ['Survey', 'Dipole', 'PointDipole']


class Survey:
    """Create a survey with sources, receivers, and data.

    A survey contains all the sources with their frequencies, receivers, and
    corresponding data.

    Underlying the survey-class is an xarray, which is basically a regular
    ndarray with axis labels and more. The module `xarray` is a soft
    dependency, and has to be installed manually to use the `Survey`
    functionality.

    This class was developed with a node-based, marine CSEM survey layout in
    mind. It is therefore optimised for and mostly tested with that setup. This
    means for a number of receivers which measure for all source positions. The
    general layout of the data for such a survey is (S, R, F), where `S` is the
    number of sources, `R` the number of receivers, and `F` the number of
    frequencies::

                             f1
            Rx1 Rx2  .  RxR /   f2
           ┌───┬───┬───┬───┐   /   .
       Tx1 │   │   │   │   │──┐   /   fF
           ├───┼───┼───┼───┤  │──┐   /
       Tx2 │   │   │   │   │──┤  │──┐
           ├───┼───┼───┼───┤  │──┤  │
        .  │   │   │   │   │──┤  │──┤
           ├───┼───┼───┼───┤  │──┤  │
       TxS │   │   │   │   │──┤  │──┤
           └───┴───┴───┴───┘  │──┤  │
              └───┴───┴───┴───┘  │──┤
                 └───┴───┴───┴───┘  │
                    └───┴───┴───┴───┘

    However, the class can also be used for a CSEM streamer-style survey
    layout (by setting `fixed=True`), where there is a moving source with one
    or several receivers at a fixed offset. The layout of the data is then also
    (S, R, F), but here `S` is the number of locations of the only source, `R`
    is the number of receiver-offsets, and `F` is the number of frequencies::

                                        f1
                Offs1     .   OffsR    /   .
              ┌─────────┬───┬─────────┐   /   fF
       TxPos1 │ Rx1-TP1 │ . │ RxR-TP1 │──┐   /
              ├─────────┼───┼─────────┤  │──┐
       TxPos2 │ Rx1-TP2 │ . │ RxR-TP2 │──┤  │
              ├─────────┼───┼─────────┤  │──┤
        .     │ .       │ . │ .       │──┤  │
              ├─────────┼───┼─────────┤  │──┤
       TxPosS │ Rx1-TPS │ . │ RxR-TPS │──┤  │
              └─────────┴───┴─────────┘  │──┤
                 └─────────┴───┴─────────┘  │
                    └─────────┴───┴─────────┘

    This means that even though there is only one source, there are actually
    `S` source dipoles, as each position is treated as a different dipole. The
    number of receiver dipoles in this case is `SxR`. This setup can also be
    used for airborne EM.


    .. warning::

        The Survey class has mainly been used and tested with ``fixed=False``,
        which is the default. It is very likely that there are things which are
        not yet implemented for ``fixed=True``. Please do report problems if
        you encounter them.


    Parameters
    ----------
    name : str
        Name of the survey

    sources, receivers : tuple, list, or dict
        Sources and receivers.

        - Tuples: Coordinates in one of the two following formats:

          - `(x, y, z, azimuth, dip)` [m, m, m, °, °];
          - `(x0, x1, y0, y1, z0, z1)` [m, m, m, m, m, m].

          Dimensions will be expanded (hence, if `n` dipoles, each parameter
          must have length 1 or `n`). These dipoles will be named sequential
          with `Tx###` and `Rx###`.

          The tuple can additionally contain an additional element at the end
          (after `dip` or `z1`), `electric`, a boolean of length 1 or `n`, that
          indicates if the dipoles are electric or magnetic.

        - List: A list of :class:`Dipole`-instances. The names of all dipoles
          in the list must be unique.

        - Dictionary: A dict of de-serialized :class:`Dipole`-instances; mainly
          used for loading from file.

    frequencies : ndarray
        Source frequencies (Hz).

    data : ndarray or None
        The observed data (dtype=np.complex128); must have shape (nsrc, nrec,
        nfreq) or, if `fixed=True`, (nsrc, noff, nfreq). If None, it will be
        initiated with NaN's.

    fixed : bool
        Node-based CSEM survey (`fixed=False`; default) or streamer-type CSEM
        survey (`fixed=True`). In the streamer-type survey, the number of
        `receivers` supplied must be a multiple of the source positions.
        In this case, the receivers are grouped into offsets.

    noise_floor, relative_error : float
        Noise floor and relative error of the data. Default to None.
        See :attr:`Survey.standard_deviation` for more info.

    std : ndarray or None
        Standard deviation of the data, same shape as data. Default to None.
        See :attr:`Survey.standard_deviation` for more info.

    """
    # Currently, `surveys.data` contains an :class:`xarray.Dataset`. As such,
    # the `Survey`-Class has an xarray-dataset as one of its attributes.
    # Probably there would be a cleaner way to simply use xarray instead of a
    # dedicated `Survey`-Class by utilizing, e.g.,
    # :func:`xarray.register_dataset_accessor`.

    def __init__(self, name, sources, receivers, frequencies, data=None,
                 fixed=0, **kwargs):
        """Initiate a new Survey instance."""

        # Store survey name and fixed.
        self.name = name
        self.fixed = fixed

        # Initiate sources.
        self._sources = self._dipole_info_to_dict(sources, 'source')

        # Initiate receivers.
        self._receivers = self._dipole_info_to_dict(receivers, 'receiver')

        # Initiate frequencies.
        self._frequencies = np.array(frequencies, dtype=np.float64, ndmin=1)

        # Initialize xarray dataset.
        self._initiate_dataset(data)

        # Get the optional keywords related to standard deviation.
        self.noise_floor = kwargs.pop('noise_floor', None)
        self.relative_error = kwargs.pop('relative_error', None)
        self.standard_deviation = kwargs.pop('std', None)

        # Ensure no kwargs left.
        if kwargs:
            raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}")

    @utils._requires('xarray')
    def _initiate_dataset(self, data):
        """Initiate Dataset; wrapped in fct to check if xarray is installed."""

        # Initialize NaN-data if not provided.
        if data is None:
            data = np.ones((len(self._sources), len(self._receivers),
                            self._frequencies.size),
                           dtype=np.complex128)*np.nan
        else:
            data = np.atleast_3d(data)

        # Create Dataset.
        self._data = xarray.Dataset(
            {'observed': xarray.DataArray(data, dims=('src', 'rec', 'freq'))},
            coords={'src': list(self.sources.keys()),
                    'rec': list(self.receivers.keys()),
                    'freq': list(self.frequencies)},
        )

        # Add attributes.
        self._data.src.attrs['long_name'] = 'Source dipole'
        self._data.src.attrs['src-dipoles'] = self.sources
        self._data.rec.attrs['long_name'] = 'Receiver dipole'
        self._data.rec.attrs['rec-dipoles'] = self.receivers
        self._data.freq.attrs['long_name'] = 'Source frequency'
        self._data.freq.attrs['units'] = 'Hz'

    def __repr__(self):
        return (f"{self.__class__.__name__}: {self.name}\n\n"
                f"{self.data.__repr__()}")

    def _repr_html_(self):
        return (f"<h4>{self.__class__.__name__}: {self.name}</h4><br>"
                f"{self.data._repr_html_()}")

    def copy(self):
        """Return a copy of the Survey."""
        return self.from_dict(self.to_dict(True))

    def to_dict(self, copy=False):
        """Store the necessary information of the Survey in a dict."""

        out = {'name': self.name, '__class__': self.__class__.__name__}

        # Add sources.
        out['sources'] = {k: v.to_dict() for k, v in self.sources.items()}

        # Add receivers.
        if self.fixed:
            rec = {}
            for key, value in self.receivers.items():
                rec[key] = {k: v.to_dict() for k, v in value.items()}
        else:
            rec = {k: v.to_dict() for k, v in self.receivers.items()}
        out['receivers'] = rec

        # Add frequencies.
        out['frequencies'] = self.frequencies

        # Add data.
        out['data'] = {}
        for key in self.data.keys():
            out['data'][key] = self.data[key].data

        # Add `noise_floor` and `relative error`.
        out['noise_floor'] = self.data.noise_floor
        out['relative_error'] = self.data.relative_error

        # Fixed.
        out['fixed'] = int(self.fixed)

        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    @utils._requires('xarray')
    def from_dict(cls, inp):
        """Convert dictionary into :class:`Survey` instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`Survey.to_dict`.
            The dictionary needs the keys `name`, `sources`, `receivers`
            `frequencies`, `data`, and `fixed`.

        Returns
        -------
        obj : :class:`Survey` instance

        """
        try:
            # Initiate survey.
            out = cls(name=inp['name'], sources=inp['sources'],
                      receivers=inp['receivers'],
                      frequencies=inp['frequencies'],
                      fixed=bool(inp['fixed']))

            # Backwards compatibility (emg3d < 0.13); remove eventually.
            if 'observed' in inp.keys():
                inp['data'] = {'observed': inp['observed']}
                warnings.warn("Survey-dict is outdated; store with new "
                              "version of `emg3d`.", FutureWarning)

            # Add all data.
            for key, value in inp['data'].items():
                out._data[key] = out.data.observed*np.nan
                out._data[key][...] = value

            # v0.14.0 onwards.
            if 'noise_floor' in inp.keys():
                out.noise_floor = inp['noise_floor']
                out.relative_error = inp['relative_error']

            return out

        except KeyError as e:
            raise KeyError(f"Variable {e} missing in `inp`.") from e

    def to_file(self, fname, name='survey', **kwargs):
        """Store Survey to a file.

        Parameters
        ----------
        fname : str
            File name inclusive ending, which defines the used data format.
            Implemented are currently:

            - `.h5` (default): Uses `h5py` to store inputs to a hierarchical,
              compressed binary hdf5 file. Recommended file format, but
              requires the module `h5py`. Default format if ending is not
              provided or not recognized.
            - `.npz`: Uses `numpy` to store inputs to a flat, compressed binary
              file. Default format if `h5py` is not installed.
            - `.json`: Uses `json` to store inputs to a hierarchical, plain
              text file.

        name : str
            Name under which the survey is stored within the file.

        kwargs : Keyword arguments, optional
            Passed through to :func:`io.save`.

        """
        from emg3d import io
        kwargs[name] = self                # Add survey to dict.
        kwargs['collect_classes'] = False  # Ensure classes are not collected.
        return io.save(fname, **kwargs)

    @classmethod
    @utils._requires('xarray')
    def from_file(cls, fname, name='survey', **kwargs):
        """Load Survey from a file.

        Parameters
        ----------
        fname : str
            File name including extension. Used backend depends on the file
            extensions:

            - '.npz': numpy-binary
            - '.h5': h5py-binary (needs `h5py`)
            - '.json': json

        name : str
            Name under which the survey is stored within the file.

        kwargs : Keyword arguments, optional
            Passed through to :func:`io.load`.


        Returns
        -------
        survey : :class:`Survey`
            The survey that was stored in the file.

        """
        from emg3d import io
        out = io.load(fname, **kwargs)
        if 'verb' in kwargs and kwargs['verb'] < 0:
            return out[0][name], out[1]
        else:
            return out[name]

    def select(self, sources=None, receivers=None, frequencies=None):
        """Return a survey with selectod sources, receivers, and frequencies.


        Parameters
        ----------
        sources, receivers, frequencies : list
            Lists containing the wanted sources, receivers, and frequencies.


        Returns
        -------
        subsurvey : :class:`emg3d.surveys.Survey`
            Survey with selected data.

        """

        # Get a dict of the survey
        survey = self.to_dict()
        isrc, irec, ifreq = slice(None), slice(None), slice(None)

        # Replace noise floor and relative error.
        noise_floor = np.atleast_3d(self.noise_floor)
        relative_error = np.atleast_3d(self.relative_error)

        # Select sources.
        if sources is not None:
            if isinstance(sources, str):
                sources = [sources, ]
            isrc = [list(self.sources).index(s) for s in sources]
            survey['sources'] = {s: survey['sources'][s] for s in sources}

            # Adjust noise floor and relative error.
            if noise_floor.shape[0] > 1:
                noise_floor = noise_floor[isrc, :, :]
            if relative_error.shape[0] > 1:
                relative_error = relative_error[isrc, :, :]

        # Select receivers.
        if receivers is not None:
            if isinstance(receivers, str):
                receivers = [receivers, ]
            irec = [list(self.receivers).index(r) for r in receivers]
            survey['receivers'] = {
                    r: survey['receivers'][r] for r in receivers}

            # Adjust noise floor and relative error.
            if noise_floor.shape[1] > 1:
                noise_floor = noise_floor[:, irec, :]
            if relative_error.shape[1] > 1:
                relative_error = relative_error[:, irec, :]

        # Select frequencies.
        if frequencies is not None:
            ifreq = np.isin(self.frequencies, frequencies)
            survey['frequencies'] = self.frequencies[ifreq]

            # Adjust noise floor and relative error.
            if noise_floor.shape[2] > 1:
                noise_floor = noise_floor[:, :, ifreq]
            if relative_error.shape[2] > 1:
                relative_error = relative_error[:, :, ifreq]

        # Replace data with selected data.
        for key in survey['data'].keys():
            data = self.data[key].data[isrc, :, :][:, irec, :][:, :, ifreq]
            survey['data'][key] = data
        survey['noise_floor'] = noise_floor
        survey['relative_error'] = relative_error

        # Return reduced survey.
        return Survey.from_dict(survey)

    @property
    def shape(self):
        """Return nsrc x nrec x nfreq.

        Note that not all source-receiver-frequency pairs do actually have
        data. Check `size` to see how many data points there are.
        """
        return self.data.observed.shape

    @property
    def size(self):
        """Return actual data size (does NOT equal nsrc x nrec x nfreq)."""
        return int(self.data.observed.count())

    @property
    def data(self):
        """Data, a :class:`xarray.DataSet` instance.

        Contains the :class:`xarray.DataArray` element `.observed`, but other
        data can be added. E.g., :class:`emg3d.simulations.Simulation` adds the
        `synthetic` array.
        """
        return self._data

    @property
    def sources(self):
        """Source dict containing all source dipoles."""
        return self._sources

    @property
    def receivers(self):
        """Receiver dict containing all receiver dipoles."""
        return self._receivers

    @property
    def src_coords(self):
        """Return source coordinates.

        The returned format is `(x, y, z, azm, dip)`, a tuple of 5 tuples.
        """

        return tuple(np.array([[s.xco, s.yco, s.zco, s.azm, s.dip] for s
                     in self.sources.values()]).T)

    @property
    def rec_coords(self):
        """Return receiver coordinates.

        The returned format is `(x, y, z, azm, dip)`, a tuple of 5 tuples. If
        `fixed=True` it returns a dict with the offsets as keys, and for each
        offset it returns the corresponding receiver coordinates as just
        outlined.
        """

        # Get receiver coordinates depending if fixed or not.
        if self.fixed:
            coords = {}
            for src in self.sources.keys():
                coords[src] = tuple(
                        np.array([[self.receivers[off][src].xco,
                                   self.receivers[off][src].yco,
                                   self.receivers[off][src].zco,
                                   self.receivers[off][src].azm,
                                   self.receivers[off][src].dip]
                                  for off in self.receivers.keys()]).T)
        else:
            coords = tuple(np.array([[r.xco, r.yco, r.zco, r.azm, r.dip] for r
                                     in self.receivers.values()]).T)

        return coords

    @property
    def rec_types(self):
        """Return receiver flags if electric, as tuple.

        Corresponds to ``rec_coords``.
        """

        # Get receiver coordinates depending if fixed or not.
        if self.fixed:
            types = {}
            for src in self.sources.keys():
                types[src] = tuple(
                        [self.receivers[off][src].electric
                         for off in list(self.receivers)])
        else:
            types = tuple([r.electric for r in self.receivers.values()])

        return types

    @property
    def frequencies(self):
        """Frequency array."""
        return self._frequencies

    @property
    def observed(self):
        r"""Returns the observed data."""
        return self._data.observed

    @observed.setter
    def observed(self, observed):
        """Update observed data."""
        self._data['observed'][...] = observed

    @property
    def standard_deviation(self):
        r"""Returns the standard deviation of the data.

        The standard deviation can be set by providing an array of the same
        dimension as the data itself:

        .. code-block:: python

            survey.standard_deviation = ndarray  # (nsrc, nrec, nfreq)

        Alternatively, one can set the `noise_floor` :math:`\epsilon_\text{nf}`
        and the `relative_error` :math:`\epsilon_\text{r}`:

        .. code-block:: python

            survey.noise_floor = float or ndarray      # (> 0 or None)
            survey.relative error = float or ndarray   # (> 0 or None)

        They must be either floats, or three-dimensional arrays of shape
        ``([nsrc or 1], [nrec or 1], [nfreq or 1])``; dimensions of one will be
        broadcasted to the corresponding size. E.g., for a dataset of arbitrary
        amount of sources and receivers with three frequencies you can define
        a purely frequency-dependent relative error via
        ``relative_error=np.array([err_f1, err_f2, err_f3])[None, None, :]``.

        The standard deviation :math:`\varsigma_i` of observation :math:`d_i`
        is then given in terms of the noise floor
        :math:`\epsilon_{\text{nf};i}` and the relative error
        :math:`\epsilon_{\text{re};i}` by

        .. math::
            :label: std

            \varsigma_i = \sqrt{
                \epsilon_{\text{nf}; i}^2 +
                \left(\epsilon_{\text{re}; i}|d_i|\right)^2 } \, .

        Note that a set standard deviation is prioritized over potentially also
        defined noise floor and relative error. To use the noise floor and the
        relative error after defining standard deviation directly you would
        have to reset it like

        .. code-block:: python

            survey.standard_deviation = None

        after which Equation :eq:`std` would be used again.

        """
        # If `std` was set, return it, else compute it from noise_floor and
        # relative_error.
        if 'std' in self._data.keys():
            return self.data['std']

        elif self.noise_floor is not None or self.relative_error is not None:

            # Initiate std (xarray of same type as the observed data)
            std = self.data.observed.copy(data=np.zeros(self.shape))

            # Add noise floor if given.
            if self.noise_floor is not None:
                std += self.noise_floor**2

            # Add relative error if given.
            if self.relative_error is not None:
                std += np.abs(self.relative_error*self.data.observed)**2

            # Return.
            return np.sqrt(std)

        else:
            # If nothing is defined, return None
            return None

    @standard_deviation.setter
    def standard_deviation(self, std):
        """Update standard deviation."""
        # If None it means basically to delete it; otherwise set it.
        if std is None and 'std' in self.data:
            del self._data['std']
        elif std is not None:
            # Ensure all values are bigger than zero.
            if np.any(std <= 0.0):
                raise ValueError(
                    "All values of `std` must be bigger than zero.")
            self._data['std'] = self.data.observed.copy(data=std)

    @property
    def noise_floor(self):
        r"""Returns the noise floor of the data.

        See :attr:`emg3d.surveys.Survey.standard_deviation` for more info.

        """
        return self.data.noise_floor

    @noise_floor.setter
    def noise_floor(self, noise_floor):
        """Update noise floor.

        See :attr:`Survey.standard_deviation` for more info.
        """
        if noise_floor is not None:

            # Ensure all values are bigger than zero.
            if np.any(noise_floor <= 0.0):
                raise ValueError(
                    "All values of `noise_floor` must be bigger than zero.")

            # Ensure it can be broadcast to data.
            try:
                _ = np.ones(self.shape)*noise_floor
            except ValueError as e:
                raise ValueError(
                    "Shape of `noise_floor` is not broadcastable to data.\n"
                    f"Shape of `noise_floor`: {np.shape(noise_floor)}; "
                    f"`data`: {self.shape}."
                ) from e

        self._data.attrs['noise_floor'] = noise_floor

    @property
    def relative_error(self):
        r"""Returns the relative error of the data.

        See :attr:`emg3d.surveys.Survey.standard_deviation` for more info.

        """
        return self.data.relative_error

    @relative_error.setter
    def relative_error(self, relative_error):
        """Update relative error.

        See :attr:`Survey.standard_deviation` for more info.
        """
        if relative_error is not None:

            # Ensure all values are bigger than zero.
            if np.any(relative_error <= 0.0):
                raise ValueError(
                    "All values of `relative_error` must be bigger than zero.")

            # Ensure it can be broadcast to data.
            try:
                _ = np.ones(self.shape)*relative_error
            except ValueError as e:
                raise ValueError(
                    "Shape of `relative_error` is not broadcastable to data.\n"
                    f"Shape of `relative_error`: {np.shape(relative_error)}; "
                    f"`data`: {self.shape}."
                ) from e

        self._data.attrs['relative_error'] = relative_error

    def _dipole_info_to_dict(self, inp, name):
        """Create dict with provided source/receiver information."""

        # Create dict depending if `inp` is list, tuple, or dict.
        if isinstance(inp, list):  # List of Dipoles

            if self.fixed and name == 'receiver':  # Streamer-type receivers.

                # Get dimensions.
                nd = len(inp)
                ns = len(self.sources)  # Number of source position.
                nr = nd//ns             # Number of receivers / source.
                dnr = len(str(nr-1))    # Max number of digits; rec.

                # Create name lists.
                rec_names = [f"{i:0{dnr}d}" for i in range(nr)]
                src_names = list(self.sources.keys())

                # Ensure receivers are multiples of source positions.
                if nd % ns != 0:
                    raise ValueError(
                            "For fixed surveys, the number of receivers\n"
                            "must be a multiple of number of sources.\n"
                            f"Provided: #src: {ns}; #rec: {nd}.")

                # Assemble dict.
                out = {'Off'+name: {} for name in rec_names}
                for i, key in enumerate(out.keys()):
                    for ii, src_name in enumerate(src_names):
                        out[key][src_name] = inp[ii + i*ns]

            else:

                out = {d.name: d for d in inp}

                # Ensure that all names were unique:
                if len(out) != len(inp):
                    raise ValueError(
                            f"There are duplicate {name} names.\n"
                            f"Provided {name}s: {len(inp)}; "
                            f"unique names: {len(out)}.")

        elif isinstance(inp, tuple):  # Tuple with coordinates

            # See if last tuple element is boolean, hence el/mag-flag.
            if isinstance(inp[-1], (list, tuple, np.ndarray)):
                provided_elmag = isinstance(inp[-1][0], (bool, np.bool_))
            else:
                provided_elmag = isinstance(inp[-1], (bool, np.bool_))

            # Get max dimension.
            nd = max([np.array(n, ndmin=1).size for n in inp])

            # Expand coordinates.
            coo = np.array([nd*[val, ] if np.array(val).size == 1 else
                           val for val in inp], dtype=np.float64)

            # Extract el/mag flag or set to ones (electric) if not provided.
            if provided_elmag:
                elmag = coo[-1, :]
                coo = coo[:-1, :]
            else:
                elmag = np.ones(nd)

            # Create dipole names (number-strings).
            prefix = 'Tx' if name == 'source' else 'Rx'
            dnd = len(str(nd-1))  # Max number of digits.
            names = [f"{prefix}{i:0{dnd}d}" for i in range(nd)]

            # Create Dipole-dict.
            if self.fixed and name == 'receiver':  # Streamer-type receivers.

                # Get dimensions.
                ns = len(self.sources)  # Number of source position.
                nr = nd//ns             # Number of receivers / source.
                dnr = len(str(nr-1))    # Max number of digits; rec.

                # Create name lists.
                rec_names = [f"{i:0{dnr}d}" for i in range(nr)]
                src_names = list(self.sources.keys())

                # Ensure receivers are multiples of source positions.
                if nd % ns != 0:
                    raise ValueError(
                            "For fixed surveys, the number of receivers\n"
                            "must be a multiple of number of sources.\n"
                            f"Provided: #src: {ns}; #rec: {nd}.")

                # Assemble dict.
                out = {'Off'+rec_name: {} for rec_name in rec_names}
                for i, key in enumerate(out.keys()):
                    for ii, src_name in enumerate(src_names):
                        iii = ii + i*ns
                        out[key][src_name] = Dipole(
                                names[iii], coo[:, iii], elmag[iii])

            else:  # Default node-type src-rec comb. and src for streamer-type.
                out = {names[i]: Dipole(names[i], coo[:, i], elmag[i])
                       for i in range(nd)}

        elif isinstance(inp, dict):  # Dict of de-serialized Dipoles.
            if self.fixed and name == 'receiver':
                out = {}
                for k, v in inp.items():
                    out[k] = {k2: Dipole.from_dict(v2) for k2, v2 in v.items()}
            else:
                out = {k: Dipole.from_dict(v) for k, v in inp.items()}

        else:
            raise TypeError(
                    f"Input format of <{name}s> not recognized: {type(inp)}.")

        return out


# # Sources and Receivers # #
@dataclass(order=True, unsafe_hash=True)
class PointDipole:
    """Infinitesimal small electric or magnetic point dipole.

    Defined by its coordinates (xco, yco, zco), its azimuth (azm), its dip, and
    its type (electric).

    Not meant to be used directly. Use :class:`Dipole` instead.


    Parameters
    ----------
    name : str
        Dipole name.

    xco, yco, zco : float
        x-, y-, and z-coordinates (m).

    azm, dip : float
        Angles (in degrees °); coordinate system is right-handed with positive
        z up; East-North-Depth:

        - azimuth (°): horizontal deviation from x-axis, anti-clockwise.
        - +/-dip (°): vertical deviation from xy-plane down/up-wards.

    electric : bool
        Electric dipole if True, magnetic dipole otherwise. Default is True.

    """
    __slots__ = ['name', 'xco', 'yco', 'zco', 'azm', 'dip', 'electric']
    name: str
    xco: float
    yco: float
    zco: float
    azm: float
    dip: float
    electric: bool


class Dipole(PointDipole):
    """Finite length dipole or point dipole.

    Expansion of the basic :class:`PointDipole` to allow for finite length
    dipoles, and to provide coordinate inputs in the form of
    (x, y, z, azimuth, dip) or (x0, x1, y0, y1, z0, z1).

    Adds attributes `is_finite`, `electrode1`, `electrode2`, `length`, and
    `coordinates` to the class.

    For *point dipoles*, this gives it a length of unity (1 m), takes its
    coordinates as center, and computes the two electrode positions.

    For *finite length dipoles* it sets the coordinates to its center and
    computes its length, azimuth, and dip.

    Finite length dipoles and point dipoles have therefore the exactly same
    signature, and can only be distinguished by the attribute `is_finite`.


    Parameters
    ----------
    name : str
        Dipole name.

    coordinates : tuple of floats
        Source coordinates, one of the following:

        - (x0, x1, y0, y1, z0, z1): finite length dipole,
        - (x, y, z, azimuth, dip): point dipole.

        The coordinates x, y, and z are in meters (m), the azimuth and dip in
        degree (°).

        Angles (coordinate system is right-handed with positive z up;
        East-North-Depth):

        - azimuth (°): horizontal deviation from x-axis, anti-clockwise.
        - +/-dip (°): vertical deviation from xy-plane down/up-wards.

    electric : bool
        Electric dipole if True, magnetic dipole otherwise. Default is True.

    """
    # These are the only kwargs that do not raise a warning.
    # These are also the only ones which are (de-)serialized.
    accepted_keys = ['strength', ]

    def __init__(self, name, coordinates, electric=True, **kwargs):
        """Check coordinates and kwargs."""

        # Add additional info to the dipole.
        for key in self.accepted_keys:
            if key in kwargs:
                setattr(self, key, kwargs.pop(key))
        if kwargs:
            raise TypeError(f"Unexpected **kwargs: {list(kwargs.keys())}")

        # Conversion to float-array fails if there are lists and tuples within
        # the tuple, or similar. This should also catch many wrong inputs.
        coords = np.array(coordinates, dtype=np.float64)

        # Check size => finite or point dipole?
        if coords.size == 5:
            self.is_finite = False

        elif coords.size == 6:
            self.is_finite = True

            # Ensure the two poles are distinct.
            if np.allclose(coords[::2], coords[1::2]):
                raise ValueError(
                        "The two poles are identical, use the format\n"
                        "(x, y, z, azimuth, dip) instead.\n"
                        f"Provided coordinates: {coordinates}.")

        else:
            raise ValueError(
                    "Dipole coordinates are wrong defined. They must be\n"
                    "defined either as a point, (x, y, z, azimuth, dip), or\n"
                    "as two poles, (x0, x1, y0, y1, z0, z1), all floats.\n"
                    f"Provided coordinates: {coordinates}.")

        # Angles: Very small angles are set to zero, because, e.g.,
        #         cos(pi/2) is roughly 6.12e-17, not 0.

        # Get xco, yco, zco, azm, and dip.
        if self.is_finite:

            # Get the two separate electrodes.
            self.electrode1 = tuple(coords[::2])
            self.electrode2 = tuple(coords[1::2])

            # Compute center.
            xco, yco, zco = np.sum(coords.reshape(3, -1), 1)/2

            # Get lengths in each direction.
            dx, dy, dz = np.diff(coords.reshape(3, -1)).ravel()

            # Length of bipole.
            self.length = np.linalg.norm([dx, dy, dz], axis=0)

            # Horizontal deviation from x-axis.
            azm = np.round(np.rad2deg(np.arctan2(dy, dx)), 5)

            # Vertical deviation from xy-plane down.
            dip = np.round(np.rad2deg(np.pi/2-np.arccos(dz/self.length)), 5)

        else:
            # Get coordinates, angles, and set length.
            xco, yco, zco, azm, dip = tuple(coords)
            self.length = 1.0

            # Get lengths in each direction (total length is 1).
            dx = np.round(np.cos(np.deg2rad(azm))*np.cos(np.deg2rad(dip)), 5)
            dy = np.round(np.sin(np.deg2rad(azm))*np.cos(np.deg2rad(dip)), 5)
            dz = np.round(np.sin(np.deg2rad(dip)), 5)

            # Get the two separate electrodes.
            self.electrode1 = (xco-dx/2, yco-dy/2, zco-dz/2)
            self.electrode2 = (xco+dx/2, yco+dy/2, zco+dz/2)

        super().__init__(name, xco, yco, zco, azm, dip, bool(electric))

    def __repr__(self):
        return (f"{self.__class__.__name__}({self.name}, "
                f"{['H', 'E'][self.electric]}, "
                f"{{{self.xco:,.1f}m; {self.yco:,.1f}m; {self.zco:,.1f}m}}, "
                f"θ={self.azm:.1f}°, φ={self.dip:.1f}°, "
                f"l={self.length:,.1f}m)")

    @property
    def coordinates(self):
        """Return coordinates.

        Returns
        -------
        coords : tuple
            Coordinates in the format (x, y, z, azimuth, dip) or (x0, x1, y0,
            y1, z0, z1). This format is used in many other routines.
        """
        if self.is_finite:
            return (self.electrode1[0], self.electrode2[0],
                    self.electrode1[1], self.electrode2[1],
                    self.electrode1[2], self.electrode2[2])
        else:
            return (self.xco, self.yco, self.zco, self.azm, self.dip)

    def copy(self):
        """Return a copy of the Dipole."""
        return self.from_dict(self.to_dict(True))

    def to_dict(self, copy=False):
        """Store the necessary information of the Dipole in a dict."""
        out = {'name': self.name, 'coordinates': self.coordinates,
               'electric': self.electric, '__class__': self.__class__.__name__}

        # Add accepted kwargs.
        for key in self.accepted_keys:
            if hasattr(self, key):
                out[key] = getattr(self, key)

        if copy:
            return deepcopy(out)
        else:
            return out

    @classmethod
    def from_dict(cls, inp):
        """Convert dictionary into :class:`Dipole` instance.

        Parameters
        ----------
        inp : dict
            Dictionary as obtained from :func:`Dipole.to_dict`. The dictionary
            needs the keys `name`, `coordinates`, and `electric`.

        Returns
        -------
        obj : :class:`Dipole` instance

        """
        try:
            kwargs = {k: v for k, v in inp.items() if k in cls.accepted_keys}
            return cls(name=inp['name'], coordinates=inp['coordinates'],
                       electric=inp['electric'], **kwargs)
        except KeyError as e:
            raise KeyError(f"Variable {e} missing in `inp`.") from e



"""
Parser for the configuration file of the command-line interface.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import os
import configparser
from pathlib import Path


def parse_config_file(args_dict):
    """Read and parse the configuration file and set defaults.

    Parameters
    ----------
    args_dict : dict
        Arguments from terminal, see :func:`emg3d.cli.main`.


    Returns
    -------
    conf : dict
        Configuration-dict.

    """

    # # Read parameter-file into dict # #
    config = args_dict.pop('config')
    configfile = os.path.abspath(config)
    cfg = configparser.ConfigParser(inline_comment_prefixes='#')

    # Check if configfile is actually a file.
    if os.path.isfile(configfile):

        # If it is, read it.
        with open(configfile) as f:
            cfg.read_file(f)

    elif config == '.':

        # If config == '.', no parameter file is given. Workaround to suppress
        # warning when one does not want to provide a config file.
        configfile = config

    # # Check the terminal arguments # #

    # Initiate terminal dict.
    term = {}

    # Add config-file to term.
    term['config_file'] = configfile

    # Get terminal input.
    for key in ['verbosity', 'nproc', 'dry_run']:
        term[key] = args_dict.pop(key)

    for key in ['forward', 'misfit', 'gradient']:
        function = args_dict.pop(key)
        if function:
            term['function'] = key
    if 'function' not in term.keys():
        term['function'] = 'forward'

    # Get file names.
    for key in ['path', 'survey', 'model', 'output']:
        term[key] = args_dict.pop(key)

    # Ensure no keys are left.
    if args_dict:
        raise TypeError(f"Unexpected parameter in **args_dict: "
                        f"{list(args_dict.keys())}")

    # Enforce some limits.
    term['verbosity'] = int(min(max(term['verbosity'], -1), 2))  # [-1, 2]
    if term['nproc'] is not None:
        term['nproc'] = int(max(term['nproc'], 1))               # [1, inf]

    # # Check file-paths and files # #

    # Check if parameter-file has a files-section, add it otherwise.
    if 'files' not in cfg.sections():
        cfg.add_section('files')

    # Get file names.
    all_files = dict(cfg.items('files'))

    # First path.
    path = term.pop('path')
    if path is None:
        path = all_files.pop('path', '.')
    path = os.path.abspath(path)

    # Initiate files dict with defaults.
    files = {'survey': 'survey', 'model': 'model', 'output': 'emg3d_out'}
    for key, value in files.items():

        config_or_default = all_files.pop(key, value)

        # Get terminal input.
        fname = term.pop(key)

        # If there was no terminal input, get config-file; else, default.
        if fname is None:
            fname = config_or_default

        # Get absolute paths.
        ffile = Path(os.path.join(path, fname))

        # Ensure there is a file ending, if not, fall back to h5p
        if ffile.suffix not in ['.h5', '.json', '.npz']:
            ffile = ffile.with_suffix('.h5')

        # Store in dict.
        files[key] = ffile

    # Ensure files and directory exist:
    for key in ['survey', 'model']:
        files[key] = str(files[key])
    files['log'] = str(files['output'].with_suffix('.log'))
    files['output'] = str(files['output'])

    # Store options (get it with getboolean, and remove from dict).
    files['store_simulation'] = cfg.getboolean(
            'files', 'store_simulation', fallback=False)
    _ = all_files.pop('store_simulation', None)

    # Ensure no keys are left.
    if all_files:
        raise TypeError(f"Unexpected parameter in [files]: "
                        f"{list(all_files.keys())}")

    # # Simulation parameters  # #

    # Initiate dict.
    simulation = {}

    # Check if parameter-file has a simulation-section, add it otherwise.
    if 'simulation' not in cfg.sections():
        cfg.add_section('simulation')
    all_sim = dict(cfg.items('simulation'))

    # Check max_workers.
    key = 'max_workers'
    _ = all_sim.pop(key, None)
    if term['nproc'] is not None:
        simulation[key] = term['nproc']
    elif cfg.has_option('simulation', key):
        simulation[key] = cfg.getint('simulation', key)
    del term['nproc']

    # Check gridding.
    key = 'gridding'
    if cfg.has_option('simulation', key):
        _ = all_sim.pop(key)
        simulation[key] = cfg.get('simulation', key)

    # Check name.
    key = 'name'
    if cfg.has_option('simulation', key):
        _ = all_sim.pop(key)
        simulation[key] = cfg.get('simulation', key)
    else:
        simulation[key] = "emg3d CLI run"

    key = 'min_offset'
    if cfg.has_option('simulation', key):
        _ = all_sim.pop(key)
        simulation[key] = cfg.getfloat('simulation', key)

    # Ensure no keys are left.
    if all_sim:
        raise TypeError(f"Unexpected parameter in [simulation]: "
                        f"{list(all_sim.keys())}")

    # # Solver parameters  # #

    # Check if parameter-file has a solver-section, add it otherwise.
    if 'solver_opts' in cfg.sections():

        # Initiate solver-dict.
        solver = {}

        all_solver = dict(cfg.items('solver_opts'))

        # Check for bools.
        for key in ['sslsolver', 'semicoarsening', 'linerelaxation']:
            if cfg.has_option('solver_opts', key):
                _ = all_solver.pop(key)
                solver[key] = cfg.getboolean('solver_opts', key)

        # Check for strings.
        for key in ['cycle', ]:
            if cfg.has_option('solver_opts', key):
                _ = all_solver.pop(key)
                solver[key] = cfg.get('solver_opts', key)

        # Check for floats.
        for key in ['tol', ]:
            if cfg.has_option('solver_opts', key):
                _ = all_solver.pop(key)
                solver[key] = float(cfg.get('solver_opts', key))

        # Check for ints.
        int_keys = ['verb', 'maxit', 'nu_init', 'nu_pre', 'nu_coarse',
                    'nu_post', 'clevel']
        for key in int_keys:
            if cfg.has_option('solver_opts', key):
                _ = all_solver.pop(key)
                solver[key] = cfg.getint('solver_opts', key)

        # Ensure no keys are left.
        if all_solver:
            raise TypeError(f"Unexpected parameter in [solver_opts]: "
                            f"{list(all_solver.keys())}")

        # Add to simulation dict if not empty.
        if solver:
            simulation['solver_opts'] = solver

    # # Data selection parameters  # #

    data = {}
    if 'data' in cfg.sections():
        # Get all parameters.
        all_data = dict(cfg.items('data'))

        for key in ['sources', 'receivers', 'frequencies']:
            value = all_data.pop(key, False)
            if value:
                if key == 'frequencies':
                    data[key] = [float(v) for v in value.split(',')]
                else:
                    data[key] = [v.strip() for v in value.split(',')]

        # Ensure no keys are left.
        if all_data:
            raise TypeError(f"Unexpected parameter in [data]: "
                            f"{list(all_data.keys())}")

    # # Gridding # #

    if 'gridding_opts' in cfg.sections():
        grid = {}

        all_grid = dict(cfg.items('gridding_opts'))

        # Check for lists.
        list_keys = ['properties', 'center', 'cell_number', 'min_width_pps',
                     'expand']
        for key in list_keys:
            if cfg.has_option('gridding_opts', key):
                _ = all_grid.pop(key)
                grid[key] = [float(v) for v in
                             cfg.get('gridding_opts', key).split(',')]

        # Check for list of lists.
        for key in ['domain', 'distance', 'stretching', 'min_width_limits']:
            if cfg.has_option('gridding_opts', key):
                _ = all_grid.pop(key)
                out = []
                for p in cfg.get('gridding_opts', key).split(';'):
                    if 'none' in p.lower():
                        out.append(None)
                    else:
                        out.append([float(v) for v in p.split(',')])
                if len(out) == 1:
                    out = out[0]
                else:
                    out = tuple(out)
                grid[key] = out

        # Check for strings.
        for key in ['mapping', 'vector']:
            if cfg.has_option('gridding_opts', key):
                _ = all_grid.pop(key)
                grid[key] = cfg.get('gridding_opts', key)

        # Check for floats.
        for key in ['frequency', 'seasurface', 'max_buffer', 'lambda_factor']:
            if cfg.has_option('gridding_opts', key):
                _ = all_grid.pop(key)
                grid[key] = float(cfg.get('gridding_opts', key))

        # Check for ints.
        for key in ['verb', ]:
            if cfg.has_option('gridding_opts', key):
                _ = all_grid.pop(key)
                grid[key] = cfg.getint('gridding_opts', key)

        # Check for bools.
        for key in ['lambda_from_center', ]:
            if cfg.has_option('gridding_opts', key):
                _ = all_grid.pop(key)
                grid[key] = cfg.getboolean('gridding_opts', key)

        # Ensure no keys are left.
        if all_grid:
            raise TypeError(f"Unexpected parameter in [gridding_opts]: "
                            f"{list(all_grid.keys())}")

        # Add to simulation dict if not empty.
        if grid:
            simulation['gridding_opts'] = grid

    # Return.
    out = {'files': files, 'simulation_options': simulation, 'data': data}
    return out, term



"""
Entry point for the command-line interface (CLI).
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import os
import sys
import argparse

from emg3d import utils
from emg3d.cli import run


def main(args=None):
    """Parsing command line inputs of CLI interface."""

    # If not explicitly called, catch arguments.
    if args is None:
        args = sys.argv[1:]

    # Start CLI-arg-parser and define arguments.
    parser = argparse.ArgumentParser(
        description="emg3d is a multigrid solver for 3D EM diffusion "
                    "with tri-axial anisotropy."
    )

    # arg: Optional parameter-file name.
    parser.add_argument(
        "config",
        nargs="?",
        default="emg3d.cfg",
        type=str,
        help=("name of config file; default is 'emg3d.cfg'; consult "
              "https://emg3d.rtfd.io/en/stable/cli.html for its format")
    )

    # arg: Number of processors.
    parser.add_argument(
        "-n", "--nproc",
        type=int,
        default=None,
        help="number of processors"
    )

    # arg: What to run
    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument(
        "-f", "--forward",
        action='store_true',
        help="compute synthetic data (default)"
    )
    group1.add_argument(
        "-m", "--misfit",
        action='store_true',
        help="compute synthetic data and their misfit"
    )
    group1.add_argument(
        "-g", "--gradient",
        action='store_true',
        help="compute synthetic data, misfit, and its gradient"
    )

    # arg: Path to files
    parser.add_argument(
        "--path",
        type=str,
        default=None,
        help="path (abs or rel); file names are relative to path"
    )

    # arg: Survey file name; relative to path
    parser.add_argument(
        "--survey",
        type=str,
        default=None,
        help="input survey file name; default is 'survey.h5'"
    )

    # arg: Model file name; relative to path
    parser.add_argument(
        "--model",
        type=str,
        default=None,
        help="input model file name; default is 'model.h5'"
    )

    # arg: Output base name; relative to path
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="output files base name; default is 'emg3d_out'"
    )

    # arg: Verbosity.
    group3 = parser.add_mutually_exclusive_group()
    group3.add_argument(
        "--verbosity",
        type=int,
        default=0,
        choices=[-1, 0, 1, 2],
        help="set verbosity; default is 0"
    )
    group3.add_argument(
        "-v", "--verbose",
        action="count",
        dest="verbosity",
        help="increase verbosity; can be used multiple times"
    )
    group3.add_argument(
        "-q", "--quiet",
        action="store_const",
        const=-1,
        dest="verbosity",
        help="decrease verbosity"
    )

    # arg: Run without emg3d-computation.
    parser.add_argument(
        "-d", "--dry-run",
        action="store_true",
        default=False,
        help="only display what would have been done"
    )

    # arg: Report
    parser.add_argument(
        "--report",
        action="store_true",
        default=False,
        help="only display emg3d report"
    )

    # arg: Version
    parser.add_argument(
        "--version",
        action="store_true",
        default=False,
        help="only display emg3d version"
    )

    # Get command line arguments.
    args_dict = vars(parser.parse_args(args))

    # Exits without simulation.
    if args_dict.pop('version'):  # emg3d version info.

        print(f"emg3d v{utils.__version__}")
        return

    elif args_dict.pop('report'):  # emg3d report.
        print(utils.Report())
        return

    elif len(sys.argv) == 1 and not os.path.isfile('emg3d.cfg'):

        # If no arguments provided, and ./emg3d.cfg does not exist, print info.
        print(parser.description)
        version = utils.__version__
        print(f"=> Type `emg3d --help` for more info (emg3d v{version}).")
        return

    # Run simulation with given command line inputs.
    run.simulation(args_dict)


if __name__ == "__main__":
    sys.exit(main())



"""
Interpolation routines mapping grids to grids, grids to fields, and fields to
grids.
"""
# Copyright 2018-2021 The emg3d Developers.
#
# This file is part of emg3d.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.  You may obtain a copy
# of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations under
# the License.

import numba as nb
import numpy as np
from scipy import interpolate, ndimage

__all__ = ['grid2grid', 'interp3d', 'MapConductivity', 'MapLgConductivity',
           'MapLnConductivity', 'MapResistivity', 'MapLgResistivity',
           'MapLnResistivity', 'edges2cellaverages']

# Numba-settings
_numba_setting = {'nogil': True, 'fastmath': True, 'cache': True}


# INTERPOLATIONS
def grid2grid(grid, values, new_grid, method='linear', extrapolate=True,
              log=False):
    """Interpolate `values` located on `grid` to `new_grid`.

    **Note 1:**
    The default method is 'linear', because it works with fields and model
    parameters. However, recommended are 'volume' for model parameters and
    'cubic' for fields.

    **Note 2:**
    For model parameters with `method='volume'` the result is quite different
    if you provide resistivity, conductivity, or the logarithm of any of the
    two. The recommended way is to provide the logarithm of resistivity or
    conductivity, in which case the output of one is indeed the inverse of the
    output of the other.


    Parameters
    ----------
    grid, new_grid : TensorMesh
        Input and output model grids;
        :class:`TensorMesh` instances.

    values : ndarray
        Model parameters; :class:`emg3d.fields.Field` instance, or a particular
        field (e.g. field.fx). For fields the method cannot be 'volume'.

    method : {<'linear'>, 'volume', 'cubic'}, optional
        The method of interpolation to perform. The volume averaging method
        ensures that the total sum of the property stays constant.

        Volume averaging is only implemented for model parameters, not for
        fields. The method 'cubic' requires at least three points in any
        direction, otherwise it will fall back to 'linear'.

        Default is 'linear', because it works with fields and model parameters.
        However, recommended are 'volume' for model parameters and 'cubic' for
        fields.

    extrapolate : bool
        If True, points on `new_grid` which are outside of `grid` are filled by
        the nearest value (if ``method='cubic'``) or by extrapolation (if
        ``method='linear'``). If False, points outside are set to zero.

        For ``method='volume'`` it always uses the nearest value for points
        outside of `grid`.

        Default is True.

    log : bool
        If True, the interpolation is carried out on a log10-scale; hence the
        same as ``10**grid2grid(grid, np.log10(values), ...)``.
        Default is False.


    Returns
    -------
    new_values : ndarray
        Values corresponding to `new_grid`.


    See Also
    --------
    get_receiver : Interpolation of model parameters or fields to (x, y, z).

    """

    # If values is a Field instance, call it recursively for each field.
    if hasattr(values, 'field') and values.field.ndim == 1:
        fx = grid2grid(grid, np.asarray(values.fx), new_grid, method,
                       extrapolate, log)
        fy = grid2grid(grid, np.asarray(values.fy), new_grid, method,
                       extrapolate, log)
        fz = grid2grid(grid, np.asarray(values.fz), new_grid, method,
                       extrapolate, log)

        # Return a field instance.
        return values.__class__(fx, fy, fz)

    # If values is a particular field, ensure method is not 'volume'.
    if not np.all(grid.vnC == values.shape) and method == 'volume':
        raise ValueError("``method='volume'`` not implemented for fields.")

    if method == 'volume':
        points = (grid.nodes_x, grid.nodes_y, grid.nodes_z)
        new_points = (new_grid.nodes_x, new_grid.nodes_y, new_grid.nodes_z)
        new_values = np.zeros(new_grid.vnC, dtype=values.dtype)
        vol = new_grid.cell_volumes.reshape(new_grid.vnC, order='F')

        # Get values from `volume_average`.
        if log:
            volume_average(*points, np.log10(values), *new_points,
                           new_values, vol)
        else:
            volume_average(*points, values, *new_points, new_values, vol)

    else:
        # Get the vectors corresponding to input data.
        points = tuple()
        new_points = tuple()
        shape = tuple()
        for i, coord in enumerate(['x', 'y', 'z']):
            if values.shape[i] == getattr(grid, 'shape_nodes')[i]:
                pts = getattr(grid, 'nodes_'+coord)
                new_pts = getattr(new_grid, 'nodes_'+coord)
            else:
                pts = getattr(grid, 'cell_centers_'+coord)
                new_pts = getattr(new_grid, 'cell_centers_'+coord)

            # Add to points.
            points += (pts, )
            new_points += (new_pts, )
            shape += (len(new_pts), )

        # Format the output points.
        xx, yy, zz = np.broadcast_arrays(
                new_points[0][:, None, None],
                new_points[1][:, None],
                new_points[2])
        new_points = np.r_[xx.ravel('F'), yy.ravel('F'), zz.ravel('F')]
        new_points = new_points.reshape(-1, 3, order='F')

        # Get values from `interp3d`.
        if extrapolate:
            fill_value = None
            mode = 'nearest'
        else:
            fill_value = 0.0
            mode = 'constant'

        if log:
            new_values = interp3d(points, np.log10(values), new_points,
                                  method, fill_value, mode)
        else:
            new_values = interp3d(points, values, new_points, method,
                                  fill_value, mode)

        new_values = new_values.reshape(shape, order='F')

    if log:
        return 10**new_values
    else:
        return new_values


def interp3d(points, values, new_points, method, fill_value, mode, cval=0.0):
    """Interpolate values in 3D either linearly or with a cubic spline.

    Return `values` corresponding to a regular 3D grid defined by `points` on
    `new_points`.

    This is a modified version of :func:`scipy.interpolate.interpn`, using
    :class:`scipy.interpolate.RegularGridInterpolator` if ``method='linear'``
    and a custom-wrapped version of :func:`scipy.ndimage.map_coordinates` if
    ``method='cubic'``. If speed is important then choose 'linear', as it can
    be significantly faster.


    Parameters
    ----------
    points : tuple of ndarray of float, with shapes ((nx, ), (ny, ) (nz, ))
        The points defining the regular grid in three dimensions.

    values : array_like, shape (nx, ny, nz)
        The data on the regular grid in three dimensions.

    new_points : tuple (rec_x, rec_y, rec_z)
        Coordinates (x, y, z) of new points.

    method : {'cubic', 'linear'}, optional
        The method of interpolation to perform, 'linear' or 'cubic'. Default is
        'cubic' (forced to 'linear' if there are less than 3 points in any
        direction).

    fill_value : float or None
        Passed to :class:`scipy.interpolate.RegularGridInterpolator` if
        ``method='linear'``: The value to use for points outside of the
        interpolation domain. If None, values outside the domain are
        extrapolated.

    mode : {'constant', 'nearest', 'mirror', 'reflect', 'wrap'}
        Passed to :func:`scipy.ndimage.map_coordinates` if ``method='cubic'``:
        Determines how the input array is extended beyond its boundaries.

    cval : float or np.nan
        Passed to :func:`scipy.ndimage.map_coordinates` if ``method='cubic'``:
        it is the value to fill past edges of input if ``mode='constant'``.
        Default is 0.0

    Returns
    -------
    new_values : ndarray
        Values corresponding to `new_points`.

    """

    # We need at least 3 points in each direction for cubic spline. This should
    # never be an issue for a realistic 3D model.
    for pts in points:
        if len(pts) < 4:
            method = 'linear'

    # Interpolation.
    if method == "linear":
        ifn = interpolate.RegularGridInterpolator(
                points=points, values=values, method="linear",
                bounds_error=False, fill_value=fill_value)

        new_values = ifn(xi=new_points)

    else:

        # Replicate the same expansion of xi as used in
        # RegularGridInterpolator, so the input xi can be quite flexible.
        xi = interpolate.interpnd._ndim_coords_from_arrays(new_points, ndim=3)
        xi_shape = xi.shape
        xi = xi.reshape(-1, 3)

        # map_coordinates uses the indices of the input data (values in this
        # case) as coordinates. We have therefore to transform our desired
        # output coordinates to this artificial coordinate system too.
        coords = np.empty(xi.T.shape)
        for i in range(3):
            coords[i] = interpolate.interp1d(
                    points[i], np.arange(len(points[i])), kind='cubic',
                    bounds_error=False, fill_value='extrapolate',)(xi[:, i])

        # map_coordinates only works for real data; split it up if complex.
        params3d = {'order': 3, 'mode': mode, 'cval': cval}
        if 'complex' in values.dtype.name:
            real = ndimage.map_coordinates(values.real, coords, **params3d)
            imag = ndimage.map_coordinates(values.imag, coords, **params3d)
            result = real + 1j*imag
        else:
            result = ndimage.map_coordinates(values, coords, **params3d)

        new_values = result.reshape(xi_shape[:-1])

    return new_values


# MAPS
MAPLIST = {}


def register_map(func):
    MAPLIST[func.__name__] = func
    return func


class _Map:
    """Maps variable `x` to computational variable `σ` (conductivity)."""

    def __init__(self, description):
        """Initiate the map."""
        self.name = self.__class__.__name__[3:]
        self.description = description

    def __repr__(self):
        return (f"{self.__class__.__name__}: {self.description}\n"
                "    Maps investigation variable `x` to\n"
                "    computational variable `σ` (conductivity).")

    def forward(self, conductivity):
        """Conductivity to mapping."""
        raise NotImplementedError("Forward map not implemented.")

    def backward(self, mapped):
        """Mapping to conductivity."""
        raise NotImplementedError("Backward map not implemented.")

    def derivative_chain(self, gradient, mapped):
        """Chain rule to map gradient from conductivity to mapping space."""
        raise NotImplementedError("Derivative chain not implemented.")

    def to_dict(self):
        """Store the map name in a dict for serialization."""
        return {'name': self.name, '__class__': '_Map'}

    @classmethod
    def from_dict(cls, inp):
        """Get :class:`_Map` instance from name in dict."""
        return MAPLIST['Map'+inp['name']]()


@register_map
class MapConductivity(_Map):
    """Maps `σ` to computational variable `σ` (conductivity).

    - forward: x = σ
    - backward: σ = x

    """

    def __init__(self):
        super().__init__('conductivity')

    def forward(self, conductivity):
        return conductivity

    def backward(self, mapped):
        return mapped

    def derivative_chain(self, gradient, mapped):
        pass


@register_map
class MapLgConductivity(_Map):
    """Maps `log_10(σ)` to computational variable `σ` (conductivity).

    - forward: x = log_10(σ)
    - backward: σ = 10^x

    """

    def __init__(self):
        super().__init__('log_10(conductivity)')

    def forward(self, conductivity):
        return np.log10(conductivity)

    def backward(self, mapped):
        return 10**mapped

    def derivative_chain(self, gradient, mapped):
        gradient *= self.backward(mapped)*np.log(10)


@register_map
class MapLnConductivity(_Map):
    """Maps `log_e(σ)` to computational variable `σ` (conductivity).

    - forward: x = log_e(σ)
    - backward: σ = exp(x)

    """

    def __init__(self):
        super().__init__('log_e(conductivity)')

    def forward(self, conductivity):
        return np.log(conductivity)

    def backward(self, mapped):
        return np.exp(mapped)

    def derivative_chain(self, gradient, mapped):
        gradient *= self.backward(mapped)


@register_map
class MapResistivity(_Map):
    """Maps `ρ` to computational variable `σ` (conductivity).

    - forward: x = ρ = σ^-1
    - backward: σ = ρ^-1 = x^-1

    """

    def __init__(self):
        super().__init__('resistivity')

    def forward(self, conductivity):
        return 1.0/conductivity

    def backward(self, mapped):
        return 1.0/mapped

    def derivative_chain(self, gradient, mapped):
        gradient *= -self.backward(mapped)**2


@register_map
class MapLgResistivity(_Map):
    """Maps `log_10(ρ)` to computational variable `σ` (conductivity).

    - forward: x = log_10(ρ) = log_10(σ^-1)
    - backward: σ = ρ^-1 = 10^-x

    """

    def __init__(self):
        super().__init__('log_10(resistivity)')

    def forward(self, conductivity):
        return np.log10(1.0/conductivity)

    def backward(self, mapped):
        return 10**-mapped

    def derivative_chain(self, gradient, mapped):
        gradient *= -self.backward(mapped)*np.log(10)


@register_map
class MapLnResistivity(_Map):
    """Maps `log_e(ρ)` to computational variable `σ` (conductivity).

    - forward: x = log_e(ρ) = log_e(σ^-1)
    - backward: σ = ρ^-1 = exp(-x)

    """

    def __init__(self):
        super().__init__('log_e(resistivity)')

    def forward(self, conductivity):
        return np.log(1.0/conductivity)

    def backward(self, mapped):
        return np.exp(-mapped)

    def derivative_chain(self, gradient, mapped):
        gradient *= -self.backward(mapped)


# VOLUME AVERAGING
@nb.njit(**_numba_setting)
def volume_average(edges_x, edges_y, edges_z, values,
                   new_edges_x, new_edges_y, new_edges_z, new_values, new_vol):
    """Interpolation using the volume averaging technique.

    The result is added to new_values.

    The original implementation (see ``emg3d v0.7.1``) followed [PlDM07]_. Joe
    Capriot took that algorithm and made it much faster for implementation in
    `discretize`. The current implementation is a simplified Numba-version of
    his Cython version (the `discretize` version works for 1D, 2D, and 3D
    meshes and can also return a sparse matrix representing the operation).


    Parameters
    ----------
    edges_[x, y, z] : ndarray
        The edges in x-, y-, and z-directions for the original grid.

    values : ndarray
        Values corresponding to `grid`.

    new_edges_[x, y, z] : ndarray
        The edges in x-, y-, and z-directions for the new grid.

    new_values : ndarray
        Array where values corresponding to `new_grid` will be added.

    new_vol : ndarray
        The volumes of the `new_grid`-cells.

    """

    # Get the weights and indices for each direction.
    wx, ix_in, ix_out = _volume_average_weights(edges_x, new_edges_x)
    wy, iy_in, iy_out = _volume_average_weights(edges_y, new_edges_y)
    wz, iz_in, iz_out = _volume_average_weights(edges_z, new_edges_z)

    # Loop over the elements and sum up the contributions.
    for iz, w_z in enumerate(wz):
        izi = iz_in[iz]
        izo = iz_out[iz]
        for iy, w_y in enumerate(wy):
            iyi = iy_in[iy]
            iyo = iy_out[iy]
            w_zy = w_z*w_y
            for ix, w_x in enumerate(wx):
                ixi = ix_in[ix]
                ixo = ix_out[ix]
                new_values[ixo, iyo, izo] += w_zy*w_x*values[ixi, iyi, izi]

    # Normalize by new volume.
    new_values /= new_vol


@nb.njit(**_numba_setting)
def _volume_average_weights(x1, x2):
    """Returaveragethe weights for the volume averaging technique.


    Parameters
    ----------
    x1, x2 : ndarray
        The edges in x-, y-, or z-directions for the original (x1) and the new
        (x2) grids.


    Returns
    -------
    hs : ndarray
        Weights for the mapping of x1 to x2.

    ix1, ix2 : ndarray
        Indices to map x1 to x2.

    """
    # Fill xs with uniques and truncate.
    # Corresponds to np.unique(np.concatenate([x1, x2])).
    n1, n2 = len(x1), len(x2)
    xs = np.empty(n1 + n2)  # Pre-allocate array containing all edges.
    i1, i2, i = 0, 0, 0
    while i1 < n1 or i2 < n2:
        if i1 < n1 and i2 < n2:
            if x1[i1] < x2[i2]:
                xs[i] = x1[i1]
                i1 += 1
            elif x1[i1] > x2[i2]:
                xs[i] = x2[i2]
                i2 += 1
            else:
                xs[i] = x1[i1]
                i1 += 1
                i2 += 1
        elif i1 < n1 and i2 == n2:
            xs[i] = x1[i1]
            i1 += 1
        elif i2 < n2 and i1 == n1:
            xs[i] = x2[i2]
            i2 += 1
        i += 1

    # Get weights and indices for the two arrays.
    # - hs corresponds to np.diff(xs) where x1 and x2 overlap; zero outside.
    # - x1[ix1] can be mapped to x2[ix2] with the corresponding weight.
    nh = i-1
    hs = np.empty(nh)                   # Pre-allocate weights.
    ix1 = np.zeros(nh, dtype=np.int32)  # Pre-allocate indices for x1.
    ix2 = np.zeros(nh, dtype=np.int32)  # Pre-allocate indices for x2.
    center = 0.0
    i1, i2, i, ii = 0, 0, 0, 0
    for i in range(nh):
        center = 0.5*(xs[i]+xs[i+1])
        if x2[0] <= center and center <= x2[n2-1]:
            hs[ii] = xs[i+1]-xs[i]
            while i1 < n1-1 and center >= x1[i1]:
                i1 += 1
            while i2 < n2-1 and center >= x2[i2]:
                i2 += 1
            ix1[ii] = min(max(i1-1, 0), n1-1)
            ix2[ii] = min(max(i2-1, 0), n2-1)
            ii += 1

    return hs[:ii], ix1[:ii], ix2[:ii]


# EDGES => CENTERS
@nb.njit(**_numba_setting)
def edges2cellaverages(ex, ey, ez, vol, out_x, out_y, out_z):
    r"""Interpolate fields defined on edges to volume-averaged cell values.


    Parameters
    ----------
    ex, ey, ez : ndarray
        Electric fields in x-, y-, and z-directions, as obtained from
        :class:`emg3d.fields.Field`.

    vol : ndarray
        Volumes of the grid, as obtained from :class:`emg3d.meshes.TensorMesh`.

    out_x, out_y, out_z : ndarray
        Arrays where the results are placed (per direction).

    """

    # Get dimensions
    nx, ny, nz = vol.shape

    # Loop over dimensions.
    for iz in range(nz+1):
        izm = max(0, iz-1)
        izp = min(nz-1, iz)

        for iy in range(ny+1):
            iym = max(0, iy-1)
            iyp = min(ny-1, iy)

            for ix in range(nx+1):
                ixm = max(0, ix-1)
                ixp = min(nx-1, ix)

                # Multiply field by volume/4.
                if ix < nx:
                    out_x[ix, iym, izm] += vol[ix, iym, izm]*ex[ix, iy, iz]/4
                    out_x[ix, iyp, izm] += vol[ix, iyp, izm]*ex[ix, iy, iz]/4
                    out_x[ix, iym, izp] += vol[ix, iym, izp]*ex[ix, iy, iz]/4
                    out_x[ix, iyp, izp] += vol[ix, iyp, izp]*ex[ix, iy, iz]/4

                if iy < ny:
                    out_y[ixm, iy, izm] += vol[ixm, iy, izm]*ey[ix, iy, iz]/4
                    out_y[ixp, iy, izm] += vol[ixp, iy, izm]*ey[ix, iy, iz]/4
                    out_y[ixm, iy, izp] += vol[ixm, iy, izp]*ey[ix, iy, iz]/4
                    out_y[ixp, iy, izp] += vol[ixp, iy, izp]*ey[ix, iy, iz]/4

                if iz < nz:
                    out_z[ixm, iym, iz] += vol[ixm, iym, iz]*ez[ix, iy, iz]/4
                    out_z[ixp, iym, iz] += vol[ixp, iym, iz]*ez[ix, iy, iz]/4
                    out_z[ixm, iyp, iz] += vol[ixm, iyp, iz]*ez[ix, iy, iz]/4
                    out_z[ixp, iyp, iz] += vol[ixp, iyp, iz]*ez[ix, iy, iz]/4
